from __future__ import annotations

import datetime
import fnmatch
import locale
import logging
import os
import pathlib
import re
import shutil
import sys
import traceback
import urllib
from collections.abc import Hashable
from contextlib import contextmanager
from functools import partial
from html import entities
from html.parser import HTMLParser
from itertools import groupby
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Generator,
    Iterable,
    Sequence,
)

import dateutil.parser
from watchfiles import Change

try:
    from zoneinfo import ZoneInfo
except ModuleNotFoundError:
    from backports.zoneinfo import ZoneInfo
import watchfiles
from markupsafe import Markup

if TYPE_CHECKING:
    from pelican.contents import Content
    from pelican.settings import Settings

logger = logging.getLogger(__name__)


def sanitised_join(base_directory: str, *parts: str) -> str:
    joined = posixize_path(os.path.abspath(os.path.join(base_directory, *parts)))
    base = posixize_path(os.path.abspath(base_directory))
    if not joined.startswith(base):
        raise RuntimeError(f"Attempted to break out of output directory to {joined}")

    return joined


def strftime(date: datetime.datetime, date_format: str) -> str:
    """
    Enhanced replacement for built-in strftime with zero stripping

    This works by 'grabbing' possible format strings (those starting with %),
    formatting them with the date, stripping any leading zeros if - prefix is
    used and replacing formatted output back.
    """

    def strip_zeros(x):
        return x.lstrip("0") or "0"

    # includes ISO date parameters added by Python 3.6
    c89_directives = "aAbBcdfGHIjmMpSUuVwWxXyYzZ%"

    # grab candidate format options
    format_options = "%[-]?."
    candidates = re.findall(format_options, date_format)

    # replace candidates with placeholders for later % formatting
    template = re.sub(format_options, "%s", date_format)

    formatted_candidates = []
    for candidate in candidates:
        # test for valid C89 directives only
        if candidate[-1] in c89_directives:
            # check for '-' prefix
            if len(candidate) == 3:  # noqa: PLR2004
                # '-' prefix
                candidate = f"%{candidate[-1]}"
                conversion = strip_zeros
            else:
                conversion = None

            # format date
            if isinstance(date, SafeDatetime):
                formatted = date.strftime(candidate, safe=False)
            else:
                formatted = date.strftime(candidate)

            # strip zeros if '-' prefix is used
            if conversion:
                formatted = conversion(formatted)
        else:
            formatted = candidate
        formatted_candidates.append(formatted)

    # put formatted candidates back and return
    return template % tuple(formatted_candidates)


class SafeDatetime(datetime.datetime):
    """Subclass of datetime that works with utf-8 format strings on PY2"""

    def strftime(self, fmt, safe=True):
        """Uses our custom strftime if supposed to be *safe*"""
        if safe:
            return strftime(self, fmt)
        else:
            return super().strftime(fmt)


class DateFormatter:
    """A date formatter object used as a jinja filter

    Uses the `strftime` implementation and makes sure jinja uses the locale
    defined in LOCALE setting
    """

    def __init__(self) -> None:
        self.locale = locale.setlocale(locale.LC_TIME)
        # python has issue with Turkish_Türkiye.1254 locale, replace it to
        # something accepted: Turkish
        if self.locale == "Turkish_Türkiye.1254":
            self.locale = "Turkish"

    def __call__(self, date: datetime.datetime, date_format: str) -> str:
        # on OSX, encoding from LC_CTYPE determines the unicode output in PY3
        # make sure it's same as LC_TIME
        with temporary_locale(self.locale, locale.LC_TIME), temporary_locale(
            self.locale, locale.LC_CTYPE
        ):
            formatted = strftime(date, date_format)

        return formatted


class memoized:
    """Function decorator to cache return values.

    If called later with the same arguments, the cached value is returned
    (not reevaluated).

    """

    def __init__(self, func: Callable) -> None:
        self.func = func
        self.cache: dict[Any, Any] = {}

    def __call__(self, *args) -> Any:
        if not isinstance(args, Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self) -> str | None:
        return self.func.__doc__

    def __get__(self, obj: Any, objtype):
        """Support instance methods."""
        fn = partial(self.__call__, obj)
        fn.cache = self.cache
        return fn


def deprecated_attribute(
    old: str,
    new: str,
    since: tuple[int, ...],
    remove: tuple[int, ...] | None = None,
    doc: str | None = None,
):
    """Attribute deprecation decorator for gentle upgrades

    For example:

        class MyClass (object):
            @deprecated_attribute(
                old='abc', new='xyz', since=(3, 2, 0), remove=(4, 1, 3))
            def abc(): return None

            def __init__(self):
                xyz = 5

    Note that the decorator needs a dummy method to attach to, but the
    content of the dummy method is ignored.
    """

    def _warn():
        version = ".".join(str(x) for x in since)
        message = [f"{old} has been deprecated since {version}"]
        if remove:
            version = ".".join(str(x) for x in remove)
            message.append(f" and will be removed by version {version}")
        message.append(f".  Use {new} instead.")
        logger.warning("".join(message))
        logger.debug("".join(str(x) for x in traceback.format_stack()))

    def fget(self):
        _warn()
        return getattr(self, new)

    def fset(self, value):
        _warn()
        setattr(self, new, value)

    def decorator(dummy):
        return property(fget=fget, fset=fset, doc=doc)

    return decorator


def get_date(string: str) -> datetime.datetime:
    """Return a datetime object from a string.

    If no format matches the given date, raise a ValueError.
    """
    string = re.sub(" +", " ", string)
    default = SafeDatetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    try:
        return dateutil.parser.parse(string, default=default)
    except (TypeError, ValueError):
        raise ValueError(f"{string!r} is not a valid date") from None


@contextmanager
def pelican_open(
    filename: str, mode: str = "r", strip_crs: bool = (sys.platform == "win32")
) -> Generator[str, None, None]:
    """Open a file and return its content"""

    # utf-8-sig will clear any BOM if present
    with open(filename, mode, encoding="utf-8-sig") as infile:
        content = infile.read()
    yield content


def slugify(
    value: str,
    regex_subs: Iterable[tuple[str, str]] = (),
    preserve_case: bool = False,
    use_unicode: bool = False,
) -> str:
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    Took from Django sources.

    For a set of sensible default regex substitutions to pass to regex_subs
    look into pelican.settings.DEFAULT_CONFIG['SLUG_REGEX_SUBSTITUTIONS'].
    """

    import unicodedata

    import unidecode

    def normalize_unicode(text: str) -> str:
        # normalize text by compatibility composition
        # see: https://en.wikipedia.org/wiki/Unicode_equivalence
        return unicodedata.normalize("NFKC", text)

    # strip tags from value
    value = Markup(value).striptags()

    # normalization
    value = normalize_unicode(value)

    if not use_unicode:
        # ASCII-fy
        value = unidecode.unidecode(value)

    # perform regex substitutions
    for src, dst in regex_subs:
        value = re.sub(
            normalize_unicode(src), normalize_unicode(dst), value, flags=re.IGNORECASE
        )

    if not preserve_case:
        value = value.lower()

    return value.strip()


def copy(source: str, destination: str, ignores: Iterable[str] | None = None) -> None:
    """Recursively copy source into destination.

    If source is a file, destination has to be a file as well.
    The function is able to copy either files or directories.

    :param source: the source file or directory
    :param destination: the destination file or directory
    :param ignores: either None, or a list of glob patterns;
        files matching those patterns will _not_ be copied.
    """

    def walk_error(err):
        logger.warning("While copying %s: %s: %s", source_, err.filename, err.strerror)

    source_ = os.path.abspath(os.path.expanduser(source))
    destination_ = os.path.abspath(os.path.expanduser(destination))

    if ignores is None:
        ignores = []

    if any(fnmatch.fnmatch(os.path.basename(source), ignore) for ignore in ignores):
        logger.info("Not copying %s due to ignores", source_)
        return

    if os.path.isfile(source_):
        dst_dir = os.path.dirname(destination_)
        if not os.path.exists(dst_dir):
            logger.info("Creating directory %s", dst_dir)
            os.makedirs(dst_dir)
        logger.info("Copying %s to %s", source_, destination_)
        copy_file(source_, destination_)

    elif os.path.isdir(source_):
        if not os.path.exists(destination_):
            logger.info("Creating directory %s", destination_)
            os.makedirs(destination_)
        if not os.path.isdir(destination_):
            logger.warning(
                "Cannot copy %s (a directory) to %s (a file)", source_, destination_
            )
            return

        for src_dir, subdirs, others in os.walk(source_, followlinks=True):
            dst_dir = os.path.join(destination_, os.path.relpath(src_dir, source_))

            subdirs[:] = (
                s for s in subdirs if not any(fnmatch.fnmatch(s, i) for i in ignores)
            )
            others[:] = (
                o for o in others if not any(fnmatch.fnmatch(o, i) for i in ignores)
            )

            if not os.path.isdir(dst_dir):
                logger.info("Creating directory %s", dst_dir)
                # Parent directories are known to exist, so 'mkdir' suffices.
                os.mkdir(dst_dir)

            for o in others:
                src_path = os.path.join(src_dir, o)
                dst_path = os.path.join(dst_dir, o)
                if os.path.isfile(src_path):
                    logger.info("Copying %s to %s", src_path, dst_path)
                    copy_file(src_path, dst_path)
                else:
                    logger.warning(
                        "Skipped copy %s (not a file or directory) to %s",
                        src_path,
                        dst_path,
                    )


def copy_file(source: str, destination: str) -> None:
    """Copy a file"""
    try:
        shutil.copyfile(source, destination)
    except OSError as e:
        logger.warning(
            "A problem occurred copying file %s to %s; %s", source, destination, e
        )


def clean_output_dir(path: str, retention: Iterable[str]) -> None:
    """Remove all files from output directory except those in retention list"""

    if not os.path.exists(path):
        logger.debug("Directory already removed: %s", path)
        return

    if not os.path.isdir(path):
        try:
            os.remove(path)
        except Exception as e:
            logger.error("Unable to delete file %s; %s", path, e)
        return

    # remove existing content from output folder unless in retention list
    for filename in os.listdir(path):
        file = os.path.join(path, filename)
        if any(filename == retain for retain in retention):
            logger.debug(
                "Skipping deletion; %s is on retention list: %s", filename, file
            )
        elif os.path.isdir(file):
            try:
                shutil.rmtree(file)
                logger.debug("Deleted directory %s", file)
            except Exception as e:
                logger.error("Unable to delete directory %s; %s", file, e)
        elif os.path.isfile(file) or os.path.islink(file):
            try:
                os.remove(file)
                logger.debug("Deleted file/link %s", file)
            except Exception as e:
                logger.error("Unable to delete file %s; %s", file, e)
        else:
            logger.error("Unable to delete %s, file type unknown", file)


def get_relative_path(path: str) -> str:
    """Return the relative path from the given path to the root path."""
    components = split_all(path)
    if components is None or len(components) <= 1:
        return os.curdir
    else:
        parents = [os.pardir] * (len(components) - 1)
        return os.path.join(*parents)


def path_to_url(path: str) -> str:
    """Return the URL corresponding to a given path."""
    if path is not None:
        path = posixize_path(path)
    return path


def posixize_path(rel_path: str) -> str:
    """Use '/' as path separator, so that source references,
    like '{static}/foo/bar.jpg' or 'extras/favicon.ico',
    will work on Windows as well as on Mac and Linux."""
    return rel_path.replace(os.sep, "/")


class _HTMLWordTruncator(HTMLParser):
    _word_regex = re.compile(
        r"{DBC}|(\w[\w'-]*)".format(
            # DBC means CJK-like characters. An character can stand for a word.
            DBC=(
                "([\u4e00-\u9fff])|"  # CJK Unified Ideographs
                "([\u3400-\u4dbf])|"  # CJK Unified Ideographs Extension A
                "([\uf900-\ufaff])|"  # CJK Compatibility Ideographs
                "([\U00020000-\U0002a6df])|"  # CJK Unified Ideographs Extension B
                "([\U0002f800-\U0002fa1f])|"  # CJK Compatibility Ideographs Supplement
                "([\u3040-\u30ff])|"  # Hiragana and Katakana
                "([\u1100-\u11ff])|"  # Hangul Jamo
                "([\uac00-\ud7ff])|"  # Hangul Compatibility Jamo
                "([\u3130-\u318f])"  # Hangul Syllables
            )
        ),
        re.UNICODE,
    )
    _word_prefix_regex = re.compile(r"\w", re.U)
    _singlets = ("br", "col", "link", "base", "img", "param", "area", "hr", "input")

    class TruncationCompleted(Exception):
        def __init__(self, truncate_at: int) -> None:
            super().__init__(truncate_at)
            self.truncate_at = truncate_at

    def __init__(self, max_words: int) -> None:
        super().__init__(convert_charrefs=False)

        self.max_words = max_words
        self.words_found = 0
        self.open_tags = []
        self.last_word_end = None
        self.truncate_at: int | None = None

    def feed(self, *args, **kwargs) -> None:
        try:
            super().feed(*args, **kwargs)
        except self.TruncationCompleted as exc:
            self.truncate_at = exc.truncate_at
        else:
            self.truncate_at = None

    def getoffset(self) -> int:
        line_start = 0
        lineno, line_offset = self.getpos()
        for _ in range(lineno - 1):
            line_start = self.rawdata.index("\n", line_start) + 1
        return line_start + line_offset

    def add_word(self, word_end: int) -> None:
        self.words_found += 1
        self.last_word_end = None
        if self.words_found == self.max_words:
            raise self.TruncationCompleted(word_end)

    def add_last_word(self) -> None:
        if self.last_word_end is not None:
            self.add_word(self.last_word_end)

    def handle_starttag(self, tag: str, attrs: Any) -> None:
        self.add_last_word()
        if tag not in self._singlets:
            self.open_tags.insert(0, tag)

    def handle_endtag(self, tag: str) -> None:
        self.add_last_word()
        try:
            i = self.open_tags.index(tag)
        except ValueError:
            pass
        else:
            # SGML: An end tag closes, back to the matching start tag,
            # all unclosed intervening start tags with omitted end tags
            del self.open_tags[: i + 1]

    def handle_data(self, data: str) -> None:
        word_end = 0
        offset = self.getoffset()

        while self.words_found < self.max_words:
            match = self._word_regex.search(data, word_end)
            if not match:
                break

            if match.start(0) > 0:
                self.add_last_word()

            word_end = match.end(0)
            self.last_word_end = offset + word_end

        if word_end < len(data):
            self.add_last_word()

    def _handle_ref(self, name: str, char: str) -> None:
        """
        Called by handle_entityref() or handle_charref() when a ref like
        `&mdash;`, `&#8212;`, or `&#x2014` is found.

        The arguments for this method are:

        - `name`: the HTML entity name (such as `mdash` or `#8212` or `#x2014`)
        - `char`: the Unicode representation of the ref (such as `—`)

        This method checks whether the entity is considered to be part of a
        word or not and, if not, signals the end of a word.
        """
        # Compute the index of the character right after the ref.
        #
        # In a string like 'prefix&mdash;suffix', the end is the sum of:
        #
        # - `self.getoffset()` (the length of `prefix`)
        # - `1` (the length of `&`)
        # - `len(name)` (the length of `mdash`)
        # - `1` (the length of `;`)
        #
        # Note that, in case of malformed HTML, the ';' character may
        # not be present.

        offset = self.getoffset()
        ref_end = offset + len(name) + 1

        try:
            if self.rawdata[ref_end] == ";":
                ref_end += 1
        except IndexError:
            # We are at the end of the string and there's no ';'
            pass

        if self.last_word_end is None:
            if self._word_prefix_regex.match(char):
                self.last_word_end = ref_end
        elif self._word_regex.match(char):
            self.last_word_end = ref_end
        else:
            self.add_last_word()

    def handle_entityref(self, name: str) -> None:
        """
        Called when an entity ref like '&mdash;' is found

        `name` is the entity ref without ampersand and semicolon (e.g. `mdash`)
        """
        try:
            codepoint = entities.name2codepoint[name]
            char = chr(codepoint)
        except KeyError:
            char = ""
        self._handle_ref(name, char)

    def handle_charref(self, name: str) -> None:
        """
        Called when a char ref like '&#8212;' or '&#x2014' is found

        `name` is the char ref without ampersand and semicolon (e.g. `#8212` or
        `#x2014`)
        """
        try:
            if name.startswith("x"):
                codepoint = int(name[1:], 16)
            else:
                codepoint = int(name)
            char = chr(codepoint)
        except (ValueError, OverflowError):
            char = ""
        self._handle_ref("#" + name, char)


def truncate_html_words(s: str, num: int, end_text: str = "…") -> str:
    """Truncates HTML to a certain number of words.

    (not counting tags and comments). Closes opened tags if they were correctly
    closed in the given html. Takes an optional argument of what should be used
    to notify that the string has been truncated, defaulting to ellipsis (…).

    Newlines in the HTML are preserved. (From the django framework).
    """
    length = int(num)
    if length <= 0:
        return ""
    truncator = _HTMLWordTruncator(length)
    truncator.feed(s)
    if truncator.truncate_at is None:
        return s
    out = s[: truncator.truncate_at]
    if end_text:
        out += " " + end_text
    # Close any tags still open
    for tag in truncator.open_tags:
        out += f"</{tag}>"
    # Return string
    return out


def truncate_html_paragraphs(s, count):
    """Truncate HTML to a certain number of paragraphs.

    :param count: number of paragraphs to keep

    Newlines in the HTML are preserved.
    """
    paragraphs = []
    tag_stop = 0
    substr = s[:]
    for _ in range(count):
        substr = substr[tag_stop:]
        tag_start = substr.find("<p>")
        tag_stop = substr.find("</p>") + len("</p>")
        paragraphs.append(substr[tag_start:tag_stop])

    return "".join(paragraphs)


def process_translations(
    content_list: list[Content],
    translation_id: str | Collection[str] | None = None,
) -> tuple[list[Content], list[Content]]:
    """Finds translations and returns them.

    For each content_list item, populates the 'translations' attribute, and
    returns a tuple with two lists (index, translations). Index list includes
    items in default language or items which have no variant in default
    language. Items with the `translation` metadata set to something else than
    `False` or `false` will be used as translations, unless all the items in
    the same group have that metadata.

    Translations and original items are determined relative to one another
    amongst items in the same group. Items are in the same group if they
    have the same value(s) for the metadata attribute(s) specified by the
    'translation_id', which must be a string or a collection of strings.
    If 'translation_id' is falsy, the identification of translations is skipped
    and all items are returned as originals.
    """

    if not translation_id:
        return content_list, []

    if isinstance(translation_id, str):
        translation_id = {translation_id}

    index = []

    try:
        content_list.sort(key=attrgetter(*translation_id))
    except TypeError:
        raise TypeError(
            f"Cannot unpack {translation_id}, 'translation_id' must be falsy, a"
            " string or a collection of strings"
        ) from None
    except AttributeError:
        raise AttributeError(
            f"Cannot use {translation_id} as 'translation_id', there "
            "appear to be items without these metadata "
            "attributes"
        ) from None

    for id_vals, items in groupby(content_list, attrgetter(*translation_id)):
        # prepare warning string
        id_vals = (id_vals,) if len(translation_id) == 1 else id_vals
        with_str = "with" + ", ".join([' {} "{{}}"'] * len(translation_id)).format(
            *translation_id
        ).format(*id_vals)

        items = list(items)
        original_items = get_original_items(items, with_str)
        index.extend(original_items)
        for a in items:
            a.translations = [x for x in items if x != a]

    translations = [x for x in content_list if x not in index]

    return index, translations


def get_original_items(items: list[Content], with_str: str) -> list[Content]:
    def _warn_source_paths(msg, items, *extra):
        args = [len(items)]
        args.extend(extra)
        args.extend(x.source_path for x in items)
        logger.warning("{}: {}".format(msg, "\n%s" * len(items)), *args)

    # warn if several items have the same lang
    for lang, lang_items in groupby(items, attrgetter("lang")):
        lang_items = list(lang_items)
        if len(lang_items) > 1:
            _warn_source_paths(
                'There are %s items "%s" with lang %s', lang_items, with_str, lang
            )

    # items with `translation` metadata will be used as translations...
    candidate_items = [
        i for i in items if i.metadata.get("translation", "false").lower() == "false"
    ]

    # ...unless all items with that slug are translations
    if not candidate_items:
        _warn_source_paths('All items ("%s") "%s" are translations', items, with_str)
        candidate_items = items

    # find items with default language
    original_items = [i for i in candidate_items if i.in_default_lang]

    # if there is no article with default language, go back one step
    if not original_items:
        original_items = candidate_items

    # warn if there are several original items
    if len(original_items) > 1:
        _warn_source_paths(
            "There are %s original (not translated) items %s", original_items, with_str
        )
    return original_items


def order_content(
    content_list: list[Content],
    order_by: str | Callable[[Content], Any] | None = "slug",
) -> list[Content]:
    """Sorts content.

    order_by can be a string of an attribute or sorting function. If order_by
    is defined, content will be ordered by that attribute or sorting function.
    By default, content is ordered by slug.

    Different content types can have default order_by attributes defined
    in settings, e.g. PAGES_ORDER_BY='sort-order', in which case `sort-order`
    should be a defined metadata attribute in each page.
    """

    if order_by:
        if callable(order_by):
            try:
                content_list.sort(key=order_by)
            except Exception:
                logger.error("Error sorting with function %s", order_by)
        elif isinstance(order_by, str):
            if order_by.startswith("reversed-"):
                order_reversed = True
                order_by = order_by.replace("reversed-", "", 1)
            else:
                order_reversed = False

            if order_by == "basename":
                content_list.sort(
                    key=lambda x: os.path.basename(x.source_path or ""),
                    reverse=order_reversed,
                )
            else:
                try:
                    content_list.sort(key=attrgetter(order_by), reverse=order_reversed)
                except AttributeError:
                    for content in content_list:
                        try:
                            getattr(content, order_by)
                        except AttributeError:
                            logger.warning(
                                'There is no "%s" attribute in "%s". '
                                "Defaulting to slug order.",
                                order_by,
                                content.get_relative_source_path(),
                                extra={
                                    "limit_msg": (
                                        "More files are missing "
                                        "the needed attribute."
                                    )
                                },
                            )
        else:
            logger.warning(
                "Invalid *_ORDER_BY setting (%s). "
                "Valid options are strings and functions.",
                order_by,
            )

    return content_list


def wait_for_changes(
    settings_file: str,
    settings: Settings,
) -> set[tuple[Change, str]]:
    content_path = settings.get("PATH", "")
    theme_path = settings.get("THEME", "")
    ignore_files = {
        fnmatch.translate(pattern) for pattern in settings.get("IGNORE_FILES", [])
    }

    candidate_paths = [
        settings_file,
        theme_path,
        content_path,
    ]

    candidate_paths.extend(
        os.path.join(content_path, path) for path in settings.get("STATIC_PATHS", [])
    )

    watching_paths = []
    for path in candidate_paths:
        if not path:
            continue
        path = os.path.abspath(path)
        if not os.path.exists(path):
            logger.warning("Unable to watch path '%s' as it does not exist.", path)
        else:
            watching_paths.append(path)

    return next(
        watchfiles.watch(
            *watching_paths,
            watch_filter=watchfiles.DefaultFilter(ignore_entity_patterns=ignore_files),  # type: ignore
            rust_timeout=0,
        )
    )


def set_date_tzinfo(
    d: datetime.datetime, tz_name: str | None = None
) -> datetime.datetime:
    """Set the timezone for dates that don't have tzinfo"""
    if tz_name and not d.tzinfo:
        timezone = ZoneInfo(tz_name)
        d = d.replace(tzinfo=timezone)
        return SafeDatetime(
            d.year, d.month, d.day, d.hour, d.minute, d.second, d.microsecond, d.tzinfo
        )
    return d


def mkdir_p(path: str) -> None:
    os.makedirs(path, exist_ok=True)


def split_all(path: str | pathlib.Path | None) -> Sequence[str] | None:
    """Split a path into a list of components

    While os.path.split() splits a single component off the back of
    `path`, this function splits all components:

    >>> split_all(os.path.join('a', 'b', 'c'))
    ['a', 'b', 'c']
    """
    if isinstance(path, str):
        components = []
        path = path.lstrip("/")
        while path:
            head, tail = os.path.split(path)
            if tail:
                components.insert(0, tail)
            elif head == path:
                components.insert(0, head)
                break
            path = head
        return components
    elif isinstance(path, pathlib.Path):
        return path.parts
    elif path is None:
        return None
    else:
        raise TypeError(
            f'"path" was {type(path)}, must be string, None, or pathlib.Path'
        )


def path_to_file_url(path: str) -> str:
    """Convert file-system path to file:// URL"""
    return urllib.parse.urljoin("file://", urllib.request.pathname2url(path))


def maybe_pluralize(count: int, singular: str, plural: str) -> str:
    """
    Returns a formatted string containing count and plural if count is not 1
    Returns count and singular if count is 1

    maybe_pluralize(0, 'Article', 'Articles') -> '0 Articles'
    maybe_pluralize(1, 'Article', 'Articles') -> '1 Article'
    maybe_pluralize(2, 'Article', 'Articles') -> '2 Articles'

    """
    selection = plural
    if count == 1:
        selection = singular
    return f"{count} {selection}"


@contextmanager
def temporary_locale(
    temp_locale: str | None = None, lc_category: int = locale.LC_ALL
) -> Generator[None, None, None]:
    """
    Enable code to run in a context with a temporary locale
    Resets the locale back when exiting context.

    Use tests.support.TestCaseWithCLocale if you want every unit test in a
    class to use the C locale.
    """
    orig_locale = locale.setlocale(lc_category)
    # python has issue with Turkish_Türkiye.1254 locale, replace it to
    # something accepted: Turkish
    if orig_locale == "Turkish_Türkiye.1254":
        orig_locale = "Turkish"
    if temp_locale:
        locale.setlocale(lc_category, temp_locale)
    yield
    locale.setlocale(lc_category, orig_locale)


def file_suffix(path: str) -> str:
    """Return the suffix of a filename in a path."""
    _, ext = os.path.splitext(os.path.basename(path))
    ret = ""
    if len(ext) > 1:
        # drop the ".", e.g., "exe", not ".exe"
        ret = ext[1:]
    return ret



import argparse
import importlib.metadata
import json
import logging
import multiprocessing
import os
import pprint
import sys
import time
import traceback
from collections.abc import Iterable

# Combines all paths to `pelican` package accessible from `sys.path`
# Makes it possible to install `pelican` and namespace plugins into different
# locations in the file system (e.g. pip with `-e` or `--user`)
from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

# pelican.log has to be the first pelican module to be loaded
# because logging.setLoggerClass has to be called before logging.getLogger
from pelican.log import console, DEFAULT_LOG_HANDLER  # noqa: I001
from pelican.log import init as init_logging
from pelican.generators import (
    ArticlesGenerator,
    PagesGenerator,
    SourceFileGenerator,
    StaticGenerator,
    TemplatePagesGenerator,
)
from pelican.plugins import signals
from pelican.plugins._utils import get_plugin_name, load_plugins
from pelican.server import ComplexHTTPRequestHandler, RootedHTTPServer
from pelican.settings import read_settings
from pelican.utils import clean_output_dir, maybe_pluralize, wait_for_changes
from pelican.writers import Writer

try:
    __version__ = importlib.metadata.version("pelican")
except Exception:
    __version__ = "unknown"

DEFAULT_CONFIG_NAME = "pelicanconf.py"
logger = logging.getLogger(__name__)


class Pelican:
    def __init__(self, settings):
        """Pelican initialization

        Performs some checks on the environment before doing anything else.
        """

        # define the default settings
        self.settings = settings

        self.path = settings["PATH"]
        self.theme = settings["THEME"]
        self.output_path = settings["OUTPUT_PATH"]
        self.ignore_files = settings["IGNORE_FILES"]
        self.delete_outputdir = settings["DELETE_OUTPUT_DIRECTORY"]
        self.output_retention = settings["OUTPUT_RETENTION"]

        self.init_path()
        self.init_plugins()
        signals.initialized.send(self)

    def init_path(self):
        if not any(p in sys.path for p in ["", os.curdir]):
            logger.debug("Adding current directory to system path")
            sys.path.insert(0, "")

    def init_plugins(self):
        self.plugins = []
        for plugin in load_plugins(self.settings):
            name = get_plugin_name(plugin)
            logger.debug("Registering plugin `%s`", name)
            try:
                plugin.register()
                self.plugins.append(plugin)
            except Exception as e:
                logger.error(
                    "Cannot register plugin `%s`\n%s",
                    name,
                    e,
                    stacklevel=2,
                )
                if self.settings.get("DEBUG", False):
                    console.print_exception()

        self.settings["PLUGINS"] = [get_plugin_name(p) for p in self.plugins]

    def run(self):
        """Run the generators and return"""
        start_time = time.time()

        context = self.settings.copy()
        # Share these among all the generators and content objects
        # They map source paths to Content objects or None
        context["generated_content"] = {}
        context["static_links"] = set()
        context["static_content"] = {}
        context["localsiteurl"] = self.settings["SITEURL"]

        generators = [
            cls(
                context=context,
                settings=self.settings,
                path=self.path,
                theme=self.theme,
                output_path=self.output_path,
            )
            for cls in self._get_generator_classes()
        ]

        # Delete the output directory if (1) the appropriate setting is True
        # and (2) that directory is not the parent of the source directory
        if self.delete_outputdir and os.path.commonpath(
            [os.path.realpath(self.output_path)]
        ) != os.path.commonpath(
            [os.path.realpath(self.output_path), os.path.realpath(self.path)]
        ):
            clean_output_dir(self.output_path, self.output_retention)

        for p in generators:
            if hasattr(p, "generate_context"):
                p.generate_context()
            if hasattr(p, "check_disabled_readers"):
                p.check_disabled_readers()

        # for plugins that create/edit the summary
        logger.debug("Signal all_generators_finalized.send(<generators>)")
        signals.all_generators_finalized.send(generators)

        # update links in the summary, etc
        for p in generators:
            if hasattr(p, "refresh_metadata_intersite_links"):
                p.refresh_metadata_intersite_links()

        writer = self._get_writer()

        for p in generators:
            if hasattr(p, "generate_output"):
                p.generate_output(writer)

        signals.finalized.send(self)

        articles_generator = next(
            g for g in generators if isinstance(g, ArticlesGenerator)
        )
        pages_generator = next(g for g in generators if isinstance(g, PagesGenerator))

        pluralized_articles = maybe_pluralize(
            (len(articles_generator.articles) + len(articles_generator.translations)),
            "article",
            "articles",
        )
        pluralized_drafts = maybe_pluralize(
            (
                len(articles_generator.drafts)
                + len(articles_generator.drafts_translations)
            ),
            "draft",
            "drafts",
        )
        pluralized_hidden_articles = maybe_pluralize(
            (
                len(articles_generator.hidden_articles)
                + len(articles_generator.hidden_translations)
            ),
            "hidden article",
            "hidden articles",
        )
        pluralized_pages = maybe_pluralize(
            (len(pages_generator.pages) + len(pages_generator.translations)),
            "page",
            "pages",
        )
        pluralized_hidden_pages = maybe_pluralize(
            (
                len(pages_generator.hidden_pages)
                + len(pages_generator.hidden_translations)
            ),
            "hidden page",
            "hidden pages",
        )
        pluralized_draft_pages = maybe_pluralize(
            (
                len(pages_generator.draft_pages)
                + len(pages_generator.draft_translations)
            ),
            "draft page",
            "draft pages",
        )

        console.print(
            f"Done: Processed {pluralized_articles}, {pluralized_drafts}, {pluralized_hidden_articles}, {pluralized_pages}, {pluralized_hidden_pages} and {pluralized_draft_pages} in {time.time() - start_time:.2f} seconds."
        )

    def _get_generator_classes(self):
        discovered_generators = [
            (ArticlesGenerator, "internal"),
            (PagesGenerator, "internal"),
        ]

        if self.settings["TEMPLATE_PAGES"]:
            discovered_generators.append((TemplatePagesGenerator, "internal"))

        if self.settings["OUTPUT_SOURCES"]:
            discovered_generators.append((SourceFileGenerator, "internal"))

        for receiver, values in signals.get_generators.send(self):
            if not isinstance(values, Iterable):
                values = (values,)
            for generator in values:
                if generator is None:
                    continue  # plugin did not return a generator
                discovered_generators.append((generator, receiver.__module__))

        # StaticGenerator must run last, so it can identify files that
        # were skipped by the other generators, and so static files can
        # have their output paths overridden by the {attach} link syntax.
        discovered_generators.append((StaticGenerator, "internal"))

        generators = []

        for generator, origin in discovered_generators:
            if not isinstance(generator, type):
                logger.error("Generator %s (%s) cannot be loaded", generator, origin)
                continue

            logger.debug("Found generator: %s (%s)", generator.__name__, origin)
            generators.append(generator)

        return generators

    def _get_writer(self):
        writers = [w for _, w in signals.get_writer.send(self) if isinstance(w, type)]
        num_writers = len(writers)

        if num_writers == 0:
            return Writer(self.output_path, settings=self.settings)

        if num_writers > 1:
            logger.warning("%s writers found, using only first one", num_writers)

        writer = writers[0]

        logger.debug("Found writer: %s (%s)", writer.__name__, writer.__module__)
        return writer(self.output_path, settings=self.settings)


class PrintSettings(argparse.Action):
    def __call__(self, parser, namespace, values, option_string):
        init_logging(name=__name__)

        try:
            instance, settings = get_instance(namespace)
        except Exception as e:
            logger.critical("%s: %s", e.__class__.__name__, e)
            console.print_exception()
            sys.exit(getattr(e, "exitcode", 1))

        if values:
            # One or more arguments provided, so only print those settings
            for setting in values:
                if setting in settings:
                    # Only add newline between setting name and value if dict
                    if isinstance(settings[setting], (dict, tuple, list)):
                        setting_format = "\n{}:\n{}"
                    else:
                        setting_format = "\n{}: {}"
                    console.print(
                        setting_format.format(
                            setting, pprint.pformat(settings[setting])
                        )
                    )
                else:
                    console.print(f"\n{setting} is not a recognized setting.")
                    break
        else:
            # No argument was given to --print-settings, so print all settings
            console.print(settings)

        parser.exit()


class ParseOverrides(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        overrides = {}
        for item in values:
            try:
                k, v = item.split("=", 1)
            except ValueError:
                raise ValueError(
                    "Extra settings must be specified as KEY=VALUE pairs "
                    f"but you specified {item}"
                ) from None
            try:
                overrides[k] = json.loads(v)
            except json.decoder.JSONDecodeError:
                raise ValueError(
                    f"Invalid JSON value: {v}. "
                    "Values specified via -e / --extra-settings flags "
                    "must be in JSON notation. "
                    "Use -e KEY='\"string\"' to specify a string value; "
                    "-e KEY=null to specify None; "
                    "-e KEY=false (or true) to specify False (or True)."
                ) from None
        setattr(namespace, self.dest, overrides)


LOG_HANDLERS = {"plain": None, "rich": DEFAULT_LOG_HANDLER}


def parse_arguments(argv=None):
    parser = argparse.ArgumentParser(
        description="A tool to generate a static blog, "
        " with restructured text input files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        dest="path",
        nargs="?",
        help="Path where to find the content files.",
        default=None,
    )

    parser.add_argument(
        "-t",
        "--theme-path",
        dest="theme",
        help="Path where to find the theme templates. If not "
        "specified, it will use the default one included with "
        "pelican.",
    )

    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        help="Where to output the generated files. If not "
        "specified, a directory will be created, named "
        '"output" in the current path.',
    )

    parser.add_argument(
        "-s",
        "--settings",
        dest="settings",
        help="The settings of the application, this is "
        f"automatically set to {DEFAULT_CONFIG_NAME} if a file exists with this "
        "name.",
    )

    parser.add_argument(
        "-d",
        "--delete-output-directory",
        dest="delete_outputdir",
        action="store_true",
        default=None,
        help="Delete the output directory.",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.INFO,
        dest="verbosity",
        help="Show all messages.",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=logging.CRITICAL,
        dest="verbosity",
        help="Show only critical errors.",
    )

    parser.add_argument(
        "-D",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="verbosity",
        help="Show all messages, including debug messages.",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the pelican version and exit.",
    )

    parser.add_argument(
        "-r",
        "--autoreload",
        dest="autoreload",
        action="store_true",
        help="Relaunch pelican each time a modification occurs"
        " on the content files.",
    )

    parser.add_argument(
        "--print-settings",
        dest="print_settings",
        nargs="*",
        action=PrintSettings,
        metavar="SETTING_NAME",
        help="Print current configuration settings and exit. "
        "Append one or more setting name arguments to see the "
        "values for specific settings only.",
    )

    parser.add_argument(
        "--relative-urls",
        dest="relative_paths",
        action="store_true",
        help="Use relative urls in output, useful for site development",
    )

    parser.add_argument(
        "--cache-path",
        dest="cache_path",
        help=(
            "Directory in which to store cache files. "
            'If not specified, defaults to "cache".'
        ),
    )

    parser.add_argument(
        "--ignore-cache",
        action="store_true",
        dest="ignore_cache",
        help="Ignore content cache from previous runs by not loading cache files.",
    )

    parser.add_argument(
        "--fatal",
        metavar="errors|warnings",
        choices=("errors", "warnings"),
        default="",
        help=(
            "Exit the program with non-zero status if any "
            "errors/warnings encountered."
        ),
    )

    parser.add_argument(
        "--log-handler",
        default="rich",
        choices=LOG_HANDLERS,
        help=(
            "Which handler to use to format log messages. "
            "The `rich` handler prints output in columns."
        ),
    )

    parser.add_argument(
        "--logs-dedup-min-level",
        default="WARNING",
        choices=("DEBUG", "INFO", "WARNING", "ERROR"),
        help=(
            "Only enable log de-duplication for levels equal"
            " to or above the specified value"
        ),
    )

    parser.add_argument(
        "-l",
        "--listen",
        dest="listen",
        action="store_true",
        help="Serve content files via HTTP and port 8000.",
    )

    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        help="Port to serve HTTP files at. (default: 8000)",
    )

    parser.add_argument(
        "-b",
        "--bind",
        dest="bind",
        help="IP to bind to when serving files via HTTP (default: 127.0.0.1)",
    )

    parser.add_argument(
        "-e",
        "--extra-settings",
        dest="overrides",
        help="Specify one or more SETTING=VALUE pairs to "
        "override settings. VALUE must be in JSON notation: "
        "specify string values as SETTING='\"some string\"'; "
        "booleans as SETTING=true or SETTING=false; "
        "None as SETTING=null.",
        nargs="*",
        action=ParseOverrides,
        default={},
    )

    args = parser.parse_args(argv)

    if args.port is not None and not args.listen:
        logger.warning("--port without --listen has no effect")
    if args.bind is not None and not args.listen:
        logger.warning("--bind without --listen has no effect")

    return args


def get_config(args):
    """Builds a config dictionary based on supplied `args`."""
    config = {}
    if args.path:
        config["PATH"] = os.path.abspath(os.path.expanduser(args.path))
    if args.output:
        config["OUTPUT_PATH"] = os.path.abspath(os.path.expanduser(args.output))
    if args.theme:
        abstheme = os.path.abspath(os.path.expanduser(args.theme))
        config["THEME"] = abstheme if os.path.exists(abstheme) else args.theme
    if args.delete_outputdir is not None:
        config["DELETE_OUTPUT_DIRECTORY"] = args.delete_outputdir
    if args.ignore_cache:
        config["LOAD_CONTENT_CACHE"] = False
    if args.cache_path:
        config["CACHE_PATH"] = args.cache_path
    if args.relative_paths:
        config["RELATIVE_URLS"] = args.relative_paths
    if args.port is not None:
        config["PORT"] = args.port
    if args.bind is not None:
        config["BIND"] = args.bind
    config["DEBUG"] = args.verbosity == logging.DEBUG
    config.update(args.overrides)

    return config


def get_instance(args):
    config_file = args.settings
    if config_file is None and os.path.isfile(DEFAULT_CONFIG_NAME):
        config_file = DEFAULT_CONFIG_NAME
        args.settings = DEFAULT_CONFIG_NAME

    settings = read_settings(config_file, override=get_config(args))

    cls = settings["PELICAN_CLASS"]
    if isinstance(cls, str):
        module, cls_name = cls.rsplit(".", 1)
        module = __import__(module)
        cls = getattr(module, cls_name)

    return cls(settings), settings


def autoreload(args, excqueue=None):
    console.print(
        "  --- AutoReload Mode: Monitoring `content`, `theme` and"
        " `settings` for changes. ---"
    )
    pelican, settings = get_instance(args)
    settings_file = os.path.abspath(args.settings)
    while True:
        try:
            pelican.run()

            changed_files = wait_for_changes(args.settings, settings)
            changed_files = {c[1] for c in changed_files}

            if settings_file in changed_files:
                pelican, settings = get_instance(args)

            console.print(
                "\n-> Modified: {}. re-generating...".format(", ".join(changed_files))
            )

        except KeyboardInterrupt:
            if excqueue is not None:
                excqueue.put(None)
                return
            raise

        except Exception as e:
            if args.verbosity == logging.DEBUG:
                if excqueue is not None:
                    excqueue.put(traceback.format_exception_only(type(e), e)[-1])
                else:
                    raise
            logger.warning(
                'Caught exception:\n"%s".', e, exc_info=settings.get("DEBUG", False)
            )


def listen(server, port, output, excqueue=None):
    # set logging level to at least "INFO" (so we can see the server requests)
    if logger.level < logging.INFO:
        logger.setLevel(logging.INFO)

    RootedHTTPServer.allow_reuse_address = True
    try:
        httpd = RootedHTTPServer(output, (server, port), ComplexHTTPRequestHandler)
    except OSError as e:
        logging.error("Could not listen on port %s, server %s.", port, server)
        if excqueue is not None:
            excqueue.put(traceback.format_exception_only(type(e), e)[-1])
        return

    try:
        console.print(f"Serving site at: http://{server}:{port} - Tap CTRL-C to stop")
        httpd.serve_forever()
    except Exception as e:
        if excqueue is not None:
            excqueue.put(traceback.format_exception_only(type(e), e)[-1])
        return

    except KeyboardInterrupt:
        httpd.socket.close()
        if excqueue is not None:
            return
        raise


def main(argv=None):
    args = parse_arguments(argv)
    logs_dedup_min_level = getattr(logging, args.logs_dedup_min_level)
    init_logging(
        level=args.verbosity,
        fatal=args.fatal,
        name=__name__,
        handler=LOG_HANDLERS[args.log_handler],
        logs_dedup_min_level=logs_dedup_min_level,
    )

    logger.debug("Pelican version: %s", __version__)
    logger.debug("Python version: %s", sys.version.split()[0])

    try:
        pelican, settings = get_instance(args)

        if args.autoreload and args.listen:
            excqueue = multiprocessing.Queue()
            p1 = multiprocessing.Process(target=autoreload, args=(args, excqueue))
            p2 = multiprocessing.Process(
                target=listen,
                args=(
                    settings.get("BIND"),
                    settings.get("PORT"),
                    settings.get("OUTPUT_PATH"),
                    excqueue,
                ),
            )
            try:
                p1.start()
                p2.start()
                exc = excqueue.get()
                if exc is not None:
                    logger.critical(exc)
            finally:
                p1.terminate()
                p2.terminate()
        elif args.autoreload:
            autoreload(args)
        elif args.listen:
            listen(
                settings.get("BIND"), settings.get("PORT"), settings.get("OUTPUT_PATH")
            )
        else:
            with console.status("Generating..."):
                pelican.run()
    except KeyboardInterrupt:
        logger.warning("Keyboard interrupt received. Exiting.")
    except Exception as e:
        logger.critical("%s: %s", e.__class__.__name__, e)

        if args.verbosity == logging.DEBUG:
            console.print_exception()
        sys.exit(getattr(e, "exitcode", 1))



import copy
import importlib.util
import inspect
import locale
import logging
import os
import re
import sys
from os.path import isabs
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, Optional

from pelican.log import LimitFilter


def load_source(name: str, path: str) -> ModuleType:
    spec = importlib.util.spec_from_file_location(name, path)
    mod = importlib.util.module_from_spec(spec)
    sys.modules[name] = mod
    spec.loader.exec_module(mod)
    return mod


logger = logging.getLogger(__name__)

Settings = Dict[str, Any]

DEFAULT_THEME = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "themes", "notmyidea"
)
DEFAULT_CONFIG = {
    "PATH": os.curdir,
    "ARTICLE_PATHS": [""],
    "ARTICLE_EXCLUDES": [],
    "PAGE_PATHS": ["pages"],
    "PAGE_EXCLUDES": [],
    "THEME": DEFAULT_THEME,
    "OUTPUT_PATH": "output",
    "READERS": {},
    "STATIC_PATHS": ["images"],
    "STATIC_EXCLUDES": [],
    "STATIC_EXCLUDE_SOURCES": True,
    "THEME_STATIC_DIR": "theme",
    "THEME_STATIC_PATHS": [
        "static",
    ],
    "FEED_ALL_ATOM": "feeds/all.atom.xml",
    "CATEGORY_FEED_ATOM": "feeds/{slug}.atom.xml",
    "AUTHOR_FEED_ATOM": "feeds/{slug}.atom.xml",
    "AUTHOR_FEED_RSS": "feeds/{slug}.rss.xml",
    "TRANSLATION_FEED_ATOM": "feeds/all-{lang}.atom.xml",
    "FEED_MAX_ITEMS": 100,
    "RSS_FEED_SUMMARY_ONLY": True,
    "FEED_APPEND_REF": False,
    "SITEURL": "",
    "SITENAME": "A Pelican Blog",
    "DISPLAY_PAGES_ON_MENU": True,
    "DISPLAY_CATEGORIES_ON_MENU": True,
    "DOCUTILS_SETTINGS": {},
    "OUTPUT_SOURCES": False,
    "OUTPUT_SOURCES_EXTENSION": ".text",
    "USE_FOLDER_AS_CATEGORY": True,
    "DEFAULT_CATEGORY": "misc",
    "WITH_FUTURE_DATES": True,
    "CSS_FILE": "main.css",
    "NEWEST_FIRST_ARCHIVES": True,
    "REVERSE_CATEGORY_ORDER": False,
    "DELETE_OUTPUT_DIRECTORY": False,
    "OUTPUT_RETENTION": [],
    "INDEX_SAVE_AS": "index.html",
    "ARTICLE_URL": "{slug}.html",
    "ARTICLE_SAVE_AS": "{slug}.html",
    "ARTICLE_ORDER_BY": "reversed-date",
    "ARTICLE_LANG_URL": "{slug}-{lang}.html",
    "ARTICLE_LANG_SAVE_AS": "{slug}-{lang}.html",
    "DRAFT_URL": "drafts/{slug}.html",
    "DRAFT_SAVE_AS": "drafts/{slug}.html",
    "DRAFT_LANG_URL": "drafts/{slug}-{lang}.html",
    "DRAFT_LANG_SAVE_AS": "drafts/{slug}-{lang}.html",
    "PAGE_URL": "pages/{slug}.html",
    "PAGE_SAVE_AS": "pages/{slug}.html",
    "PAGE_ORDER_BY": "basename",
    "PAGE_LANG_URL": "pages/{slug}-{lang}.html",
    "PAGE_LANG_SAVE_AS": "pages/{slug}-{lang}.html",
    "DRAFT_PAGE_URL": "drafts/pages/{slug}.html",
    "DRAFT_PAGE_SAVE_AS": "drafts/pages/{slug}.html",
    "DRAFT_PAGE_LANG_URL": "drafts/pages/{slug}-{lang}.html",
    "DRAFT_PAGE_LANG_SAVE_AS": "drafts/pages/{slug}-{lang}.html",
    "STATIC_URL": "{path}",
    "STATIC_SAVE_AS": "{path}",
    "STATIC_CREATE_LINKS": False,
    "STATIC_CHECK_IF_MODIFIED": False,
    "CATEGORY_URL": "category/{slug}.html",
    "CATEGORY_SAVE_AS": "category/{slug}.html",
    "TAG_URL": "tag/{slug}.html",
    "TAG_SAVE_AS": "tag/{slug}.html",
    "AUTHOR_URL": "author/{slug}.html",
    "AUTHOR_SAVE_AS": "author/{slug}.html",
    "PAGINATION_PATTERNS": [
        (1, "{name}{extension}", "{name}{extension}"),
        (2, "{name}{number}{extension}", "{name}{number}{extension}"),
    ],
    "YEAR_ARCHIVE_URL": "",
    "YEAR_ARCHIVE_SAVE_AS": "",
    "MONTH_ARCHIVE_URL": "",
    "MONTH_ARCHIVE_SAVE_AS": "",
    "DAY_ARCHIVE_URL": "",
    "DAY_ARCHIVE_SAVE_AS": "",
    "RELATIVE_URLS": False,
    "DEFAULT_LANG": "en",
    "ARTICLE_TRANSLATION_ID": "slug",
    "PAGE_TRANSLATION_ID": "slug",
    "DIRECT_TEMPLATES": ["index", "tags", "categories", "authors", "archives"],
    "THEME_TEMPLATES_OVERRIDES": [],
    "PAGINATED_TEMPLATES": {
        "index": None,
        "tag": None,
        "category": None,
        "author": None,
    },
    "PELICAN_CLASS": "pelican.Pelican",
    "DEFAULT_DATE_FORMAT": "%a %d %B %Y",
    "DATE_FORMATS": {},
    "MARKDOWN": {
        "extension_configs": {
            "markdown.extensions.codehilite": {"css_class": "highlight"},
            "markdown.extensions.extra": {},
            "markdown.extensions.meta": {},
        },
        "output_format": "html5",
    },
    "JINJA_FILTERS": {},
    "JINJA_GLOBALS": {},
    "JINJA_TESTS": {},
    "JINJA_ENVIRONMENT": {
        "trim_blocks": True,
        "lstrip_blocks": True,
        "extensions": [],
    },
    "LOG_FILTER": [],
    "LOCALE": [""],  # defaults to user locale
    "DEFAULT_PAGINATION": False,
    "DEFAULT_ORPHANS": 0,
    "DEFAULT_METADATA": {},
    "FILENAME_METADATA": r"(?P<date>\d{4}-\d{2}-\d{2}).*",
    "PATH_METADATA": "",
    "EXTRA_PATH_METADATA": {},
    "ARTICLE_PERMALINK_STRUCTURE": "",
    "TYPOGRIFY": False,
    "TYPOGRIFY_IGNORE_TAGS": [],
    "TYPOGRIFY_DASHES": "default",
    "SUMMARY_END_SUFFIX": "…",
    "SUMMARY_MAX_LENGTH": 50,
    "PLUGIN_PATHS": [],
    "PLUGINS": None,
    "PYGMENTS_RST_OPTIONS": {},
    "TEMPLATE_PAGES": {},
    "TEMPLATE_EXTENSIONS": [".html"],
    "IGNORE_FILES": [".*"],
    "SLUG_REGEX_SUBSTITUTIONS": [
        (r"[^\w\s-]", ""),  # remove non-alphabetical/whitespace/'-' chars
        (r"(?u)\A\s*", ""),  # strip leading whitespace
        (r"(?u)\s*\Z", ""),  # strip trailing whitespace
        (r"[-\s]+", "-"),  # reduce multiple whitespace or '-' to single '-'
    ],
    "INTRASITE_LINK_REGEX": "[{|](?P<what>.*?)[|}]",
    "SLUGIFY_SOURCE": "title",
    "SLUGIFY_USE_UNICODE": False,
    "SLUGIFY_PRESERVE_CASE": False,
    "CACHE_CONTENT": False,
    "CONTENT_CACHING_LAYER": "reader",
    "CACHE_PATH": "cache",
    "GZIP_CACHE": True,
    "CHECK_MODIFIED_METHOD": "mtime",
    "LOAD_CONTENT_CACHE": False,
    "FORMATTED_FIELDS": ["summary"],
    "PORT": 8000,
    "BIND": "127.0.0.1",
}

PYGMENTS_RST_OPTIONS = None


def read_settings(
    path: Optional[str] = None, override: Optional[Settings] = None
) -> Settings:
    settings = override or {}

    if path:
        settings = dict(get_settings_from_file(path), **settings)

    if settings:
        settings = handle_deprecated_settings(settings)

    if path:
        # Make relative paths absolute
        def getabs(maybe_relative, base_path=path):
            if isabs(maybe_relative):
                return maybe_relative
            return os.path.abspath(
                os.path.normpath(
                    os.path.join(os.path.dirname(base_path), maybe_relative)
                )
            )

        for p in ["PATH", "OUTPUT_PATH", "THEME", "CACHE_PATH"]:
            if settings.get(p) is not None:
                absp = getabs(settings[p])
                # THEME may be a name rather than a path
                if p != "THEME" or os.path.exists(absp):
                    settings[p] = absp

        if settings.get("PLUGIN_PATHS") is not None:
            settings["PLUGIN_PATHS"] = [
                getabs(pluginpath) for pluginpath in settings["PLUGIN_PATHS"]
            ]

    settings = dict(copy.deepcopy(DEFAULT_CONFIG), **settings)
    settings = configure_settings(settings)

    # This is because there doesn't seem to be a way to pass extra
    # parameters to docutils directive handlers, so we have to have a
    # variable here that we'll import from within Pygments.run (see
    # rstdirectives.py) to see what the user defaults were.
    global PYGMENTS_RST_OPTIONS  # noqa: PLW0603
    PYGMENTS_RST_OPTIONS = settings.get("PYGMENTS_RST_OPTIONS", None)
    return settings


def get_settings_from_module(module: Optional[ModuleType] = None) -> Settings:
    """Loads settings from a module, returns a dictionary."""

    context = {}
    if module is not None:
        context.update((k, v) for k, v in inspect.getmembers(module) if k.isupper())
    return context


def get_settings_from_file(path: str) -> Settings:
    """Loads settings from a file path, returning a dict."""

    name, ext = os.path.splitext(os.path.basename(path))
    module = load_source(name, path)
    return get_settings_from_module(module)


def get_jinja_environment(settings: Settings) -> Settings:
    """Sets the environment for Jinja"""

    jinja_env = settings.setdefault(
        "JINJA_ENVIRONMENT", DEFAULT_CONFIG["JINJA_ENVIRONMENT"]
    )

    # Make sure we include the defaults if the user has set env variables
    for key, value in DEFAULT_CONFIG["JINJA_ENVIRONMENT"].items():
        if key not in jinja_env:
            jinja_env[key] = value

    return settings


def _printf_s_to_format_field(printf_string: str, format_field: str) -> str:
    """Tries to replace %s with {format_field} in the provided printf_string.
    Raises ValueError in case of failure.
    """
    TEST_STRING = "PELICAN_PRINTF_S_DEPRECATION"
    expected = printf_string % TEST_STRING

    result = printf_string.replace("{", "{{").replace("}", "}}") % f"{{{format_field}}}"
    if result.format(**{format_field: TEST_STRING}) != expected:
        raise ValueError(f"Failed to safely replace %s with {{{format_field}}}")

    return result


def handle_deprecated_settings(settings: Settings) -> Settings:
    """Converts deprecated settings and issues warnings. Issues an exception
    if both old and new setting is specified.
    """

    # PLUGIN_PATH -> PLUGIN_PATHS
    if "PLUGIN_PATH" in settings:
        logger.warning(
            "PLUGIN_PATH setting has been replaced by "
            "PLUGIN_PATHS, moving it to the new setting name."
        )
        settings["PLUGIN_PATHS"] = settings["PLUGIN_PATH"]
        del settings["PLUGIN_PATH"]

    # PLUGIN_PATHS: str -> [str]
    if isinstance(settings.get("PLUGIN_PATHS"), str):
        logger.warning(
            "Defining PLUGIN_PATHS setting as string "
            "has been deprecated (should be a list)"
        )
        settings["PLUGIN_PATHS"] = [settings["PLUGIN_PATHS"]]

    # JINJA_EXTENSIONS -> JINJA_ENVIRONMENT > extensions
    if "JINJA_EXTENSIONS" in settings:
        logger.warning(
            "JINJA_EXTENSIONS setting has been deprecated, "
            "moving it to JINJA_ENVIRONMENT setting."
        )
        settings["JINJA_ENVIRONMENT"]["extensions"] = settings["JINJA_EXTENSIONS"]
        del settings["JINJA_EXTENSIONS"]

    # {ARTICLE,PAGE}_DIR -> {ARTICLE,PAGE}_PATHS
    for key in ["ARTICLE", "PAGE"]:
        old_key = key + "_DIR"
        new_key = key + "_PATHS"
        if old_key in settings:
            logger.warning(
                "Deprecated setting %s, moving it to %s list", old_key, new_key
            )
            settings[new_key] = [settings[old_key]]  # also make a list
            del settings[old_key]

    # EXTRA_TEMPLATES_PATHS -> THEME_TEMPLATES_OVERRIDES
    if "EXTRA_TEMPLATES_PATHS" in settings:
        logger.warning(
            "EXTRA_TEMPLATES_PATHS is deprecated use "
            "THEME_TEMPLATES_OVERRIDES instead."
        )
        if settings.get("THEME_TEMPLATES_OVERRIDES"):
            raise Exception(
                "Setting both EXTRA_TEMPLATES_PATHS and "
                "THEME_TEMPLATES_OVERRIDES is not permitted. Please move to "
                "only setting THEME_TEMPLATES_OVERRIDES."
            )
        settings["THEME_TEMPLATES_OVERRIDES"] = settings["EXTRA_TEMPLATES_PATHS"]
        del settings["EXTRA_TEMPLATES_PATHS"]

    # MD_EXTENSIONS -> MARKDOWN
    if "MD_EXTENSIONS" in settings:
        logger.warning(
            "MD_EXTENSIONS is deprecated use MARKDOWN "
            "instead. Falling back to the default."
        )
        settings["MARKDOWN"] = DEFAULT_CONFIG["MARKDOWN"]

    # LESS_GENERATOR -> Webassets plugin
    # FILES_TO_COPY -> STATIC_PATHS, EXTRA_PATH_METADATA
    for old, new, doc in [
        ("LESS_GENERATOR", "the Webassets plugin", None),
        (
            "FILES_TO_COPY",
            "STATIC_PATHS and EXTRA_PATH_METADATA",
            "https://github.com/getpelican/pelican/"
            "blob/main/docs/settings.rst#path-metadata",
        ),
    ]:
        if old in settings:
            message = f"The {old} setting has been removed in favor of {new}"
            if doc:
                message += f", see {doc} for details"
            logger.warning(message)

    # PAGINATED_DIRECT_TEMPLATES -> PAGINATED_TEMPLATES
    if "PAGINATED_DIRECT_TEMPLATES" in settings:
        message = "The {} setting has been removed in favor of {}".format(
            "PAGINATED_DIRECT_TEMPLATES", "PAGINATED_TEMPLATES"
        )
        logger.warning(message)

        # set PAGINATED_TEMPLATES
        if "PAGINATED_TEMPLATES" not in settings:
            settings["PAGINATED_TEMPLATES"] = {
                "tag": None,
                "category": None,
                "author": None,
            }

        for t in settings["PAGINATED_DIRECT_TEMPLATES"]:
            if t not in settings["PAGINATED_TEMPLATES"]:
                settings["PAGINATED_TEMPLATES"][t] = None
        del settings["PAGINATED_DIRECT_TEMPLATES"]

    # {SLUG,CATEGORY,TAG,AUTHOR}_SUBSTITUTIONS ->
    # {SLUG,CATEGORY,TAG,AUTHOR}_REGEX_SUBSTITUTIONS
    url_settings_url = "http://docs.getpelican.com/en/latest/settings.html#url-settings"
    flavours = {"SLUG", "CATEGORY", "TAG", "AUTHOR"}
    old_values = {
        f: settings[f + "_SUBSTITUTIONS"]
        for f in flavours
        if f + "_SUBSTITUTIONS" in settings
    }
    new_values = {
        f: settings[f + "_REGEX_SUBSTITUTIONS"]
        for f in flavours
        if f + "_REGEX_SUBSTITUTIONS" in settings
    }
    if old_values and new_values:
        raise Exception(
            "Setting both {new_key} and {old_key} (or variants thereof) is "
            "not permitted. Please move to only setting {new_key}.".format(
                old_key="SLUG_SUBSTITUTIONS", new_key="SLUG_REGEX_SUBSTITUTIONS"
            )
        )
    if old_values:
        message = (
            "{} and variants thereof are deprecated and will be "
            "removed in the future. Please use {} and variants thereof "
            "instead. Check {}.".format(
                "SLUG_SUBSTITUTIONS", "SLUG_REGEX_SUBSTITUTIONS", url_settings_url
            )
        )
        logger.warning(message)
        if old_values.get("SLUG"):
            for f in ("CATEGORY", "TAG"):
                if old_values.get(f):
                    old_values[f] = old_values["SLUG"] + old_values[f]
            old_values["AUTHOR"] = old_values.get("AUTHOR", [])
        for f in flavours:
            if old_values.get(f) is not None:
                regex_subs = []
                # by default will replace non-alphanum characters
                replace = True
                for tpl in old_values[f]:
                    try:
                        src, dst, skip = tpl
                        if skip:
                            replace = False
                    except ValueError:
                        src, dst = tpl
                    regex_subs.append((re.escape(src), dst.replace("\\", r"\\")))

                if replace:
                    regex_subs += [
                        (r"[^\w\s-]", ""),
                        (r"(?u)\A\s*", ""),
                        (r"(?u)\s*\Z", ""),
                        (r"[-\s]+", "-"),
                    ]
                else:
                    regex_subs += [
                        (r"(?u)\A\s*", ""),
                        (r"(?u)\s*\Z", ""),
                    ]
                settings[f + "_REGEX_SUBSTITUTIONS"] = regex_subs
            settings.pop(f + "_SUBSTITUTIONS", None)

    # `%s` -> '{slug}` or `{lang}` in FEED settings
    for key in ["TRANSLATION_FEED_ATOM", "TRANSLATION_FEED_RSS"]:
        if (
            settings.get(key)
            and not isinstance(settings[key], Path)
            and "%s" in settings[key]
        ):
            logger.warning("%%s usage in %s is deprecated, use {lang} instead.", key)
            try:
                settings[key] = _printf_s_to_format_field(settings[key], "lang")
            except ValueError:
                logger.warning(
                    "Failed to convert %%s to {lang} for %s. "
                    "Falling back to default.",
                    key,
                )
                settings[key] = DEFAULT_CONFIG[key]
    for key in [
        "AUTHOR_FEED_ATOM",
        "AUTHOR_FEED_RSS",
        "CATEGORY_FEED_ATOM",
        "CATEGORY_FEED_RSS",
        "TAG_FEED_ATOM",
        "TAG_FEED_RSS",
    ]:
        if (
            settings.get(key)
            and not isinstance(settings[key], Path)
            and "%s" in settings[key]
        ):
            logger.warning("%%s usage in %s is deprecated, use {slug} instead.", key)
            try:
                settings[key] = _printf_s_to_format_field(settings[key], "slug")
            except ValueError:
                logger.warning(
                    "Failed to convert %%s to {slug} for %s. "
                    "Falling back to default.",
                    key,
                )
                settings[key] = DEFAULT_CONFIG[key]

    # CLEAN_URLS
    if settings.get("CLEAN_URLS", False):
        logger.warning(
            "Found deprecated `CLEAN_URLS` in settings."
            " Modifying the following settings for the"
            " same behaviour."
        )

        settings["ARTICLE_URL"] = "{slug}/"
        settings["ARTICLE_LANG_URL"] = "{slug}-{lang}/"
        settings["PAGE_URL"] = "pages/{slug}/"
        settings["PAGE_LANG_URL"] = "pages/{slug}-{lang}/"

        for setting in ("ARTICLE_URL", "ARTICLE_LANG_URL", "PAGE_URL", "PAGE_LANG_URL"):
            logger.warning("%s = '%s'", setting, settings[setting])

    # AUTORELOAD_IGNORE_CACHE -> --ignore-cache
    if settings.get("AUTORELOAD_IGNORE_CACHE"):
        logger.warning(
            "Found deprecated `AUTORELOAD_IGNORE_CACHE` in "
            "settings. Use --ignore-cache instead."
        )
        settings.pop("AUTORELOAD_IGNORE_CACHE")

    # ARTICLE_PERMALINK_STRUCTURE
    if settings.get("ARTICLE_PERMALINK_STRUCTURE", False):
        logger.warning(
            "Found deprecated `ARTICLE_PERMALINK_STRUCTURE` in"
            " settings.  Modifying the following settings for"
            " the same behaviour."
        )

        structure = settings["ARTICLE_PERMALINK_STRUCTURE"]

        # Convert %(variable) into {variable}.
        structure = re.sub(r"%\((\w+)\)s", r"{\g<1>}", structure)

        # Convert %x into {date:%x} for strftime
        structure = re.sub(r"(%[A-z])", r"{date:\g<1>}", structure)

        # Strip a / prefix
        structure = re.sub("^/", "", structure)

        for setting in (
            "ARTICLE_URL",
            "ARTICLE_LANG_URL",
            "PAGE_URL",
            "PAGE_LANG_URL",
            "DRAFT_URL",
            "DRAFT_LANG_URL",
            "ARTICLE_SAVE_AS",
            "ARTICLE_LANG_SAVE_AS",
            "DRAFT_SAVE_AS",
            "DRAFT_LANG_SAVE_AS",
            "PAGE_SAVE_AS",
            "PAGE_LANG_SAVE_AS",
        ):
            settings[setting] = os.path.join(structure, settings[setting])
            logger.warning("%s = '%s'", setting, settings[setting])

    # {,TAG,CATEGORY,TRANSLATION}_FEED -> {,TAG,CATEGORY,TRANSLATION}_FEED_ATOM
    for new, old in [
        ("FEED", "FEED_ATOM"),
        ("TAG_FEED", "TAG_FEED_ATOM"),
        ("CATEGORY_FEED", "CATEGORY_FEED_ATOM"),
        ("TRANSLATION_FEED", "TRANSLATION_FEED_ATOM"),
    ]:
        if settings.get(new, False):
            logger.warning(
                "Found deprecated `%(new)s` in settings. Modify %(new)s "
                "to %(old)s in your settings and theme for the same "
                "behavior. Temporarily setting %(old)s for backwards "
                "compatibility.",
                {"new": new, "old": old},
            )
            settings[old] = settings[new]

    # Warn if removed WRITE_SELECTED is present
    if "WRITE_SELECTED" in settings:
        logger.warning(
            "WRITE_SELECTED is present in settings but this functionality was removed. "
            "It will have no effect."
        )

    return settings


def configure_settings(settings: Settings) -> Settings:
    """Provide optimizations, error checking, and warnings for the given
    settings.
    Also, specify the log messages to be ignored.
    """
    if "PATH" not in settings or not os.path.isdir(settings["PATH"]):
        raise Exception(
            "You need to specify a path containing the content"
            " (see pelican --help for more information)"
        )

    # specify the log messages to be ignored
    log_filter = settings.get("LOG_FILTER", DEFAULT_CONFIG["LOG_FILTER"])
    LimitFilter._ignore.update(set(log_filter))

    # lookup the theme in "pelican/themes" if the given one doesn't exist
    if not os.path.isdir(settings["THEME"]):
        theme_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "themes", settings["THEME"]
        )
        if os.path.exists(theme_path):
            settings["THEME"] = theme_path
        else:
            raise Exception("Could not find the theme {}".format(settings["THEME"]))

    # standardize strings to lowercase strings
    for key in ["DEFAULT_LANG"]:
        if key in settings:
            settings[key] = settings[key].lower()

    # set defaults for Jinja environment
    settings = get_jinja_environment(settings)

    # standardize strings to lists
    for key in ["LOCALE"]:
        if key in settings and isinstance(settings[key], str):
            settings[key] = [settings[key]]

    # check settings that must be a particular type
    for key, types in [
        ("OUTPUT_SOURCES_EXTENSION", str),
        ("FILENAME_METADATA", str),
    ]:
        if key in settings and not isinstance(settings[key], types):
            value = settings.pop(key)
            logger.warn(
                "Detected misconfigured %s (%s), falling back to the default (%s)",
                key,
                value,
                DEFAULT_CONFIG[key],
            )

    # try to set the different locales, fallback on the default.
    locales = settings.get("LOCALE", DEFAULT_CONFIG["LOCALE"])

    for locale_ in locales:
        try:
            locale.setlocale(locale.LC_ALL, str(locale_))
            break  # break if it is successful
        except locale.Error:
            pass
    else:
        logger.warning(
            "Locale could not be set. Check the LOCALE setting, ensuring it "
            "is valid and available on your system."
        )

    if "SITEURL" in settings:
        # If SITEURL has a trailing slash, remove it and provide a warning
        siteurl = settings["SITEURL"]
        if siteurl.endswith("/"):
            settings["SITEURL"] = siteurl[:-1]
            logger.warning("Removed extraneous trailing slash from SITEURL.")
        # If SITEURL is defined but FEED_DOMAIN isn't,
        # set FEED_DOMAIN to SITEURL
        if "FEED_DOMAIN" not in settings:
            settings["FEED_DOMAIN"] = settings["SITEURL"]

    # check content caching layer and warn of incompatibilities
    if (
        settings.get("CACHE_CONTENT", False)
        and settings.get("CONTENT_CACHING_LAYER", "") == "generator"
        and not settings.get("WITH_FUTURE_DATES", True)
    ):
        logger.warning(
            "WITH_FUTURE_DATES conflicts with CONTENT_CACHING_LAYER "
            "set to 'generator', use 'reader' layer instead"
        )

    # Warn if feeds are generated with both SITEURL & FEED_DOMAIN undefined
    feed_keys = [
        "FEED_ATOM",
        "FEED_RSS",
        "FEED_ALL_ATOM",
        "FEED_ALL_RSS",
        "CATEGORY_FEED_ATOM",
        "CATEGORY_FEED_RSS",
        "AUTHOR_FEED_ATOM",
        "AUTHOR_FEED_RSS",
        "TAG_FEED_ATOM",
        "TAG_FEED_RSS",
        "TRANSLATION_FEED_ATOM",
        "TRANSLATION_FEED_RSS",
    ]

    if any(settings.get(k) for k in feed_keys):
        if not settings.get("SITEURL"):
            logger.warning(
                "Feeds generated without SITEURL set properly may not be valid"
            )

    if "TIMEZONE" not in settings:
        logger.warning(
            "No timezone information specified in the settings. Assuming"
            " your timezone is UTC for feed generation. Check "
            "https://docs.getpelican.com/en/latest/settings.html#TIMEZONE "
            "for more information"
        )

    # fix up pagination rules
    from pelican.paginator import PaginationRule

    pagination_rules = [
        PaginationRule(*r)
        for r in settings.get(
            "PAGINATION_PATTERNS",
            DEFAULT_CONFIG["PAGINATION_PATTERNS"],
        )
    ]
    settings["PAGINATION_PATTERNS"] = sorted(
        pagination_rules,
        key=lambda r: r[0],
    )

    # Save people from accidentally setting a string rather than a list
    path_keys = (
        "ARTICLE_EXCLUDES",
        "DEFAULT_METADATA",
        "DIRECT_TEMPLATES",
        "THEME_TEMPLATES_OVERRIDES",
        "FILES_TO_COPY",
        "IGNORE_FILES",
        "PAGINATED_DIRECT_TEMPLATES",
        "PLUGINS",
        "STATIC_EXCLUDES",
        "STATIC_PATHS",
        "THEME_STATIC_PATHS",
        "ARTICLE_PATHS",
        "PAGE_PATHS",
    )
    for PATH_KEY in filter(lambda k: k in settings, path_keys):
        if isinstance(settings[PATH_KEY], str):
            logger.warning(
                "Detected misconfiguration with %s setting "
                "(must be a list), falling back to the default",
                PATH_KEY,
            )
            settings[PATH_KEY] = DEFAULT_CONFIG[PATH_KEY]

    # Add {PAGE,ARTICLE}_PATHS to {ARTICLE,PAGE}_EXCLUDES
    mutually_exclusive = ("ARTICLE", "PAGE")
    for type_1, type_2 in [mutually_exclusive, mutually_exclusive[::-1]]:
        try:
            includes = settings[type_1 + "_PATHS"]
            excludes = settings[type_2 + "_EXCLUDES"]
            for path in includes:
                if path not in excludes:
                    excludes.append(path)
        except KeyError:
            continue  # setting not specified, nothing to do

    return settings



import locale
import logging
import os
import shutil
from datetime import timezone
from sys import platform
from tempfile import mkdtemp

try:
    from zoneinfo import ZoneInfo
except ModuleNotFoundError:
    from backports.zoneinfo import ZoneInfo

from pelican import utils
from pelican.generators import TemplatePagesGenerator
from pelican.settings import read_settings
from pelican.tests.support import (
    LoggedTestCase,
    get_article,
    locale_available,
    unittest,
)
from pelican.writers import Writer


class ClassDeprAttr:
    _new_attribute = "new_value"

    @utils.deprecated_attribute(
        old="_old_attribute", new="_new_attribute", since=(3, 1, 0), remove=(4, 1, 3)
    )
    def _old_attribute():
        return None


class TestUtils(LoggedTestCase):
    def setUp(self):
        super().setUp()
        self.temp_output = mkdtemp(prefix="pelicantests.")

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.temp_output)

    def test_deprecated_attribute(self):
        test_class = ClassDeprAttr()
        value = test_class._old_attribute
        self.assertEqual(value, test_class._new_attribute)
        self.assertLogCountEqual(
            count=1,
            msg=(
                "_old_attribute has been deprecated since 3.1.0 and will be "
                "removed by version 4.1.3.  Use _new_attribute instead"
            ),
            level=logging.WARNING,
        )

    def test_get_date(self):
        # valid ones
        date = utils.SafeDatetime(year=2012, month=11, day=22)
        date_hour = utils.SafeDatetime(year=2012, month=11, day=22, hour=22, minute=11)
        date_hour_z = utils.SafeDatetime(
            year=2012, month=11, day=22, hour=22, minute=11, tzinfo=timezone.utc
        )
        date_hour_est = utils.SafeDatetime(
            year=2012, month=11, day=22, hour=22, minute=11, tzinfo=ZoneInfo("EST")
        )
        date_hour_sec = utils.SafeDatetime(
            year=2012, month=11, day=22, hour=22, minute=11, second=10
        )
        date_hour_sec_z = utils.SafeDatetime(
            year=2012,
            month=11,
            day=22,
            hour=22,
            minute=11,
            second=10,
            tzinfo=timezone.utc,
        )
        date_hour_sec_est = utils.SafeDatetime(
            year=2012,
            month=11,
            day=22,
            hour=22,
            minute=11,
            second=10,
            tzinfo=ZoneInfo("EST"),
        )
        date_hour_sec_frac_z = utils.SafeDatetime(
            year=2012,
            month=11,
            day=22,
            hour=22,
            minute=11,
            second=10,
            microsecond=123000,
            tzinfo=timezone.utc,
        )
        dates = {
            "2012-11-22": date,
            "2012/11/22": date,
            "2012-11-22 22:11": date_hour,
            "2012/11/22 22:11": date_hour,
            "22-11-2012": date,
            "22/11/2012": date,
            "22.11.2012": date,
            "22.11.2012 22:11": date_hour,
            "2012-11-22T22:11Z": date_hour_z,
            "2012-11-22T22:11-0500": date_hour_est,
            "2012-11-22 22:11:10": date_hour_sec,
            "2012-11-22T22:11:10Z": date_hour_sec_z,
            "2012-11-22T22:11:10-0500": date_hour_sec_est,
            "2012-11-22T22:11:10.123Z": date_hour_sec_frac_z,
        }

        # examples from http://www.w3.org/TR/NOTE-datetime
        iso_8601_date = utils.SafeDatetime(year=1997, month=7, day=16)
        iso_8601_date_hour_tz = utils.SafeDatetime(
            year=1997,
            month=7,
            day=16,
            hour=19,
            minute=20,
            tzinfo=ZoneInfo("Europe/London"),
        )
        iso_8601_date_hour_sec_tz = utils.SafeDatetime(
            year=1997,
            month=7,
            day=16,
            hour=19,
            minute=20,
            second=30,
            tzinfo=ZoneInfo("Europe/London"),
        )
        iso_8601_date_hour_sec_ms_tz = utils.SafeDatetime(
            year=1997,
            month=7,
            day=16,
            hour=19,
            minute=20,
            second=30,
            microsecond=450000,
            tzinfo=ZoneInfo("Europe/London"),
        )
        iso_8601 = {
            "1997-07-16": iso_8601_date,
            "1997-07-16T19:20+01:00": iso_8601_date_hour_tz,
            "1997-07-16T19:20:30+01:00": iso_8601_date_hour_sec_tz,
            "1997-07-16T19:20:30.45+01:00": iso_8601_date_hour_sec_ms_tz,
        }

        # invalid ones
        invalid_dates = ["2010-110-12", "yay"]

        for value, expected in dates.items():
            self.assertEqual(utils.get_date(value), expected, value)

        for value, expected in iso_8601.items():
            self.assertEqual(utils.get_date(value), expected, value)

        for item in invalid_dates:
            self.assertRaises(ValueError, utils.get_date, item)

    def test_slugify(self):
        samples = (
            ("this is a test", "this-is-a-test"),
            ("this        is a test", "this-is-a-test"),
            ("this → is ← a ↑ test", "this-is-a-test"),
            ("this--is---a test", "this-is-a-test"),
            (
                "unicode測試許功蓋，你看到了嗎？",
                "unicodece-shi-xu-gong-gai-ni-kan-dao-liao-ma",
            ),
            (
                "大飯原発４号機、１８日夜起動へ",
                "da-fan-yuan-fa-4hao-ji-18ri-ye-qi-dong-he",
            ),
        )

        settings = read_settings()
        subs = settings["SLUG_REGEX_SUBSTITUTIONS"]

        for value, expected in samples:
            self.assertEqual(utils.slugify(value, regex_subs=subs), expected)

        self.assertEqual(utils.slugify("Cat", regex_subs=subs), "cat")
        self.assertEqual(
            utils.slugify("Cat", regex_subs=subs, preserve_case=False), "cat"
        )
        self.assertEqual(
            utils.slugify("Cat", regex_subs=subs, preserve_case=True), "Cat"
        )

    def test_slugify_use_unicode(self):
        samples = (
            ("this is a test", "this-is-a-test"),
            ("this        is a test", "this-is-a-test"),
            ("this → is ← a ↑ test", "this-is-a-test"),
            ("this--is---a test", "this-is-a-test"),
            ("unicode測試許功蓋，你看到了嗎？", "unicode測試許功蓋你看到了嗎"),
            ("Çığ", "çığ"),
        )

        settings = read_settings()
        subs = settings["SLUG_REGEX_SUBSTITUTIONS"]

        for value, expected in samples:
            self.assertEqual(
                utils.slugify(value, regex_subs=subs, use_unicode=True), expected
            )

        # check with preserve case
        self.assertEqual(
            utils.slugify("Çığ", regex_subs=subs, preserve_case=True, use_unicode=True),
            "Çığ",
        )

        # check normalization
        samples = (
            ("大飯原発４号機、１８日夜起動へ", "大飯原発4号機18日夜起動へ"),
            (
                "\N{LATIN SMALL LETTER C}\N{COMBINING CEDILLA}",
                "\N{LATIN SMALL LETTER C WITH CEDILLA}",
            ),
        )
        for value, expected in samples:
            self.assertEqual(
                utils.slugify(value, regex_subs=subs, use_unicode=True), expected
            )

    def test_slugify_substitute(self):
        samples = (
            ("C++ is based on C", "cpp-is-based-on-c"),
            ("C+++ test C+ test", "cpp-test-c-test"),
            ("c++, c#, C#, C++", "cpp-c-sharp-c-sharp-cpp"),
            ("c++-streams", "cpp-streams"),
        )

        settings = read_settings()
        subs = [
            (r"C\+\+", "CPP"),
            (r"C#", "C-SHARP"),
        ] + settings["SLUG_REGEX_SUBSTITUTIONS"]
        for value, expected in samples:
            self.assertEqual(utils.slugify(value, regex_subs=subs), expected)

    def test_slugify_substitute_and_keeping_non_alphanum(self):
        samples = (
            ("Fedora QA", "fedora.qa"),
            ("C++ is used by Fedora QA", "cpp is used by fedora.qa"),
            ("C++ is based on C", "cpp is based on c"),
            ("C+++ test C+ test", "cpp+ test c+ test"),
        )

        subs = [
            (r"Fedora QA", "fedora.qa"),
            (r"c\+\+", "cpp"),
        ]
        for value, expected in samples:
            self.assertEqual(utils.slugify(value, regex_subs=subs), expected)

    def test_get_relative_path(self):
        samples = (
            (os.path.join("test", "test.html"), os.pardir),
            (
                os.path.join("test", "test", "test.html"),
                os.path.join(os.pardir, os.pardir),
            ),
            ("test.html", os.curdir),
            (os.path.join("/test", "test.html"), os.pardir),
            (
                os.path.join("/test", "test", "test.html"),
                os.path.join(os.pardir, os.pardir),
            ),
            ("/test.html", os.curdir),
        )

        for value, expected in samples:
            self.assertEqual(utils.get_relative_path(value), expected)

    def test_truncate_html_words(self):
        # Plain text.
        self.assertEqual(utils.truncate_html_words("short string", 20), "short string")
        self.assertEqual(
            utils.truncate_html_words("word " * 100, 20), "word " * 20 + "…"
        )

        # Plain text with Unicode content.
        self.assertEqual(
            utils.truncate_html_words(
                "我愿意这样，朋友——我独自远行，不但没有你，\
                 并且再没有别的影在黑暗里。",
                12,
            ),
            "我愿意这样，朋友——我独自远行" + " …",
        )
        self.assertEqual(
            utils.truncate_html_words(
                "Ты мелькнула, ты предстала, Снова сердце задрожало,", 3
            ),
            "Ты мелькнула, ты" + " …",
        )
        self.assertEqual(
            utils.truncate_html_words("Trong đầm gì đẹp bằng sen", 4),
            "Trong đầm gì đẹp" + " …",
        )

        # Words enclosed or intervaled by HTML tags.
        self.assertEqual(
            utils.truncate_html_words("<p>" + "word " * 100 + "</p>", 20),
            "<p>" + "word " * 20 + "…</p>",
        )
        self.assertEqual(
            utils.truncate_html_words(
                '<span\nstyle="\n…\n">' + "word " * 100 + "</span>", 20
            ),
            '<span\nstyle="\n…\n">' + "word " * 20 + "…</span>",
        )
        self.assertEqual(
            utils.truncate_html_words("<br>" + "word " * 100, 20),
            "<br>" + "word " * 20 + "…",
        )
        self.assertEqual(
            utils.truncate_html_words("<!-- comment -->" + "word " * 100, 20),
            "<!-- comment -->" + "word " * 20 + "…",
        )

        # Words enclosed or intervaled by HTML tags with a custom end
        # marker containing HTML tags.
        self.assertEqual(
            utils.truncate_html_words(
                "<p>" + "word " * 100 + "</p>", 20, "<span>marker</span>"
            ),
            "<p>" + "word " * 20 + "<span>marker</span></p>",
        )
        self.assertEqual(
            utils.truncate_html_words(
                '<span\nstyle="\n…\n">' + "word " * 100 + "</span>",
                20,
                "<span>marker</span>",
            ),
            '<span\nstyle="\n…\n">' + "word " * 20 + "<span>marker</span></span>",
        )
        self.assertEqual(
            utils.truncate_html_words(
                "<br>" + "word " * 100, 20, "<span>marker</span>"
            ),
            "<br>" + "word " * 20 + "<span>marker</span>",
        )
        self.assertEqual(
            utils.truncate_html_words(
                "<!-- comment -->" + "word " * 100, 20, "<span>marker</span>"
            ),
            "<!-- comment -->" + "word " * 20 + "<span>marker</span>",
        )

        # Words with hypens and apostrophes.
        self.assertEqual(utils.truncate_html_words("a-b " * 100, 20), "a-b " * 20 + "…")
        self.assertEqual(
            utils.truncate_html_words("it's " * 100, 20), "it's " * 20 + "…"
        )

        # Words with HTML entity references.
        self.assertEqual(
            utils.truncate_html_words("&eacute; " * 100, 20), "&eacute; " * 20 + "…"
        )
        self.assertEqual(
            utils.truncate_html_words("caf&eacute; " * 100, 20),
            "caf&eacute; " * 20 + "…",
        )
        self.assertEqual(
            utils.truncate_html_words("&egrave;lite " * 100, 20),
            "&egrave;lite " * 20 + "…",
        )
        self.assertEqual(
            utils.truncate_html_words("cafeti&eacute;re " * 100, 20),
            "cafeti&eacute;re " * 20 + "…",
        )
        self.assertEqual(
            utils.truncate_html_words("&int;dx " * 100, 20), "&int;dx " * 20 + "…"
        )

        # Words with HTML character references inside and outside
        # the ASCII range.
        self.assertEqual(
            utils.truncate_html_words("&#xe9; " * 100, 20), "&#xe9; " * 20 + "…"
        )
        self.assertEqual(
            utils.truncate_html_words("&#x222b;dx " * 100, 20), "&#x222b;dx " * 20 + "…"
        )

        # Words with invalid or broken HTML references.
        self.assertEqual(utils.truncate_html_words("&invalid;", 20), "&invalid;")
        self.assertEqual(
            utils.truncate_html_words("&#9999999999;", 20), "&#9999999999;"
        )
        self.assertEqual(
            utils.truncate_html_words("&#xfffffffff;", 20), "&#xfffffffff;"
        )
        self.assertEqual(utils.truncate_html_words("&mdash text", 20), "&mdash text")
        self.assertEqual(utils.truncate_html_words("&#1234 text", 20), "&#1234 text")
        self.assertEqual(utils.truncate_html_words("&#xabc text", 20), "&#xabc text")

    def test_truncate_html_paragraphs(self):
        one = "<p>one</p>"

        self.assertEqual(utils.truncate_html_paragraphs(one, 0), "")
        self.assertEqual(utils.truncate_html_paragraphs(one, 1), one)
        self.assertEqual(utils.truncate_html_paragraphs(one, 2), one)

        two = one + "<p>two</p>"
        self.assertEqual(utils.truncate_html_paragraphs(two, 1), one)
        self.assertEqual(utils.truncate_html_paragraphs(two, 2), two)

        three = two + "<p>three</p>"
        self.assertEqual(utils.truncate_html_paragraphs(three, 1), one)
        self.assertEqual(utils.truncate_html_paragraphs(three, 2), two)
        self.assertEqual(utils.truncate_html_paragraphs(three, 3), three)
        self.assertEqual(utils.truncate_html_paragraphs(three, 4), three)

    def test_process_translations(self):
        fr_articles = []
        en_articles = []

        # create a bunch of articles
        # 0: no translation metadata
        fr_articles.append(
            get_article(lang="fr", slug="yay0", title="Titre", content="en français")
        )
        en_articles.append(
            get_article(lang="en", slug="yay0", title="Title", content="in english")
        )
        # 1: translation metadata on default lang
        fr_articles.append(
            get_article(lang="fr", slug="yay1", title="Titre", content="en français")
        )
        en_articles.append(
            get_article(
                lang="en",
                slug="yay1",
                title="Title",
                content="in english",
                translation="true",
            )
        )
        # 2: translation metadata not on default lang
        fr_articles.append(
            get_article(
                lang="fr",
                slug="yay2",
                title="Titre",
                content="en français",
                translation="true",
            )
        )
        en_articles.append(
            get_article(lang="en", slug="yay2", title="Title", content="in english")
        )
        # 3: back to default language detection if all items have the
        #    translation metadata
        fr_articles.append(
            get_article(
                lang="fr",
                slug="yay3",
                title="Titre",
                content="en français",
                translation="yep",
            )
        )
        en_articles.append(
            get_article(
                lang="en",
                slug="yay3",
                title="Title",
                content="in english",
                translation="yes",
            )
        )
        # 4-5: translation pairs with the same slug but different category
        fr_articles.append(
            get_article(
                lang="fr",
                slug="yay4",
                title="Titre",
                content="en français",
                category="foo",
            )
        )
        en_articles.append(
            get_article(
                lang="en",
                slug="yay4",
                title="Title",
                content="in english",
                category="foo",
            )
        )
        fr_articles.append(
            get_article(
                lang="fr",
                slug="yay4",
                title="Titre",
                content="en français",
                category="bar",
            )
        )
        en_articles.append(
            get_article(
                lang="en",
                slug="yay4",
                title="Title",
                content="in english",
                category="bar",
            )
        )

        # try adding articles in both orders
        for lang0_articles, lang1_articles in (
            (fr_articles, en_articles),
            (en_articles, fr_articles),
        ):
            articles = lang0_articles + lang1_articles

            # test process_translations with falsy translation_id
            index, trans = utils.process_translations(articles, translation_id=None)
            for i in range(6):
                for lang_articles in [en_articles, fr_articles]:
                    self.assertIn(lang_articles[i], index)
                    self.assertNotIn(lang_articles[i], trans)

            # test process_translations with simple and complex translation_id
            for translation_id in ["slug", {"slug", "category"}]:
                index, trans = utils.process_translations(
                    articles, translation_id=translation_id
                )

                for a in [
                    en_articles[0],
                    fr_articles[1],
                    en_articles[2],
                    en_articles[3],
                    en_articles[4],
                    en_articles[5],
                ]:
                    self.assertIn(a, index)
                    self.assertNotIn(a, trans)

                for a in [
                    fr_articles[0],
                    en_articles[1],
                    fr_articles[2],
                    fr_articles[3],
                    fr_articles[4],
                    fr_articles[5],
                ]:
                    self.assertIn(a, trans)
                    self.assertNotIn(a, index)

                for i in range(6):
                    self.assertIn(en_articles[i], fr_articles[i].translations)
                    self.assertIn(fr_articles[i], en_articles[i].translations)

                for a_arts in [en_articles, fr_articles]:
                    for b_arts in [en_articles, fr_articles]:
                        if translation_id == "slug":
                            self.assertIn(a_arts[4], b_arts[5].translations)
                            self.assertIn(a_arts[5], b_arts[4].translations)
                        elif translation_id == {"slug", "category"}:
                            self.assertNotIn(a_arts[4], b_arts[5].translations)
                            self.assertNotIn(a_arts[5], b_arts[4].translations)

    def test_clean_output_dir(self):
        retention = ()
        test_directory = os.path.join(self.temp_output, "clean_output")
        content = os.path.join(os.path.dirname(__file__), "content")
        shutil.copytree(content, test_directory)
        utils.clean_output_dir(test_directory, retention)
        self.assertTrue(os.path.isdir(test_directory))
        self.assertListEqual([], os.listdir(test_directory))
        shutil.rmtree(test_directory)

    def test_clean_output_dir_not_there(self):
        retention = ()
        test_directory = os.path.join(self.temp_output, "does_not_exist")
        utils.clean_output_dir(test_directory, retention)
        self.assertFalse(os.path.exists(test_directory))

    def test_clean_output_dir_is_file(self):
        retention = ()
        test_directory = os.path.join(self.temp_output, "this_is_a_file")
        f = open(test_directory, "w")
        f.write("")
        f.close()
        utils.clean_output_dir(test_directory, retention)
        self.assertFalse(os.path.exists(test_directory))

    def test_strftime(self):
        d = utils.SafeDatetime(2012, 8, 29)

        # simple formatting
        self.assertEqual(utils.strftime(d, "%d/%m/%y"), "29/08/12")
        self.assertEqual(utils.strftime(d, "%d/%m/%Y"), "29/08/2012")

        # RFC 3339
        self.assertEqual(
            utils.strftime(d, "%Y-%m-%dT%H:%M:%SZ"), "2012-08-29T00:00:00Z"
        )

        # % escaped
        self.assertEqual(utils.strftime(d, "%d%%%m%%%y"), "29%08%12")
        self.assertEqual(utils.strftime(d, "%d %% %m %% %y"), "29 % 08 % 12")
        # not valid % formatter
        self.assertEqual(
            utils.strftime(d, "10% reduction in %Y"), "10% reduction in 2012"
        )
        self.assertEqual(
            utils.strftime(d, "%10 reduction in %Y"), "%10 reduction in 2012"
        )

        # with text
        self.assertEqual(
            utils.strftime(d, "Published in %d-%m-%Y"), "Published in 29-08-2012"
        )

        # with non-ascii text
        self.assertEqual(
            utils.strftime(d, "%d/%m/%Y Øl trinken beim Besäufnis"),
            "29/08/2012 Øl trinken beim Besäufnis",
        )

        # alternative formatting options
        self.assertEqual(utils.strftime(d, "%-d/%-m/%y"), "29/8/12")
        self.assertEqual(utils.strftime(d, "%-H:%-M:%-S"), "0:0:0")

        d = utils.SafeDatetime(2012, 8, 9)
        self.assertEqual(utils.strftime(d, "%-d/%-m/%y"), "9/8/12")

        d = utils.SafeDatetime(2021, 1, 8)
        self.assertEqual(utils.strftime(d, "%G - %-V - %u"), "2021 - 1 - 5")

    # test the output of utils.strftime in a different locale
    # Turkish locale
    @unittest.skipUnless(
        locale_available("tr_TR.UTF-8") or locale_available("Turkish"),
        "Turkish locale needed",
    )
    def test_strftime_locale_dependent_turkish(self):
        temp_locale = "Turkish" if platform == "win32" else "tr_TR.UTF-8"

        with utils.temporary_locale(temp_locale):
            d = utils.SafeDatetime(2012, 8, 29)

            # simple
            self.assertEqual(utils.strftime(d, "%d %B %Y"), "29 Ağustos 2012")
            self.assertEqual(
                utils.strftime(d, "%A, %d %B %Y"), "Çarşamba, 29 Ağustos 2012"
            )

            # with text
            self.assertEqual(
                utils.strftime(d, "Yayınlanma tarihi: %A, %d %B %Y"),
                "Yayınlanma tarihi: Çarşamba, 29 Ağustos 2012",
            )

            # non-ascii format candidate (someone might pass it… for some reason)
            self.assertEqual(
                utils.strftime(d, "%Y yılında %üretim artışı"),
                "2012 yılında %üretim artışı",
            )

    # test the output of utils.strftime in a different locale
    # French locale
    @unittest.skipUnless(
        locale_available("fr_FR.UTF-8") or locale_available("French"),
        "French locale needed",
    )
    def test_strftime_locale_dependent_french(self):
        temp_locale = "French" if platform == "win32" else "fr_FR.UTF-8"

        with utils.temporary_locale(temp_locale):
            d = utils.SafeDatetime(2012, 8, 29)

            # simple
            self.assertEqual(utils.strftime(d, "%d %B %Y"), "29 août 2012")

            # depending on OS, the first letter is m or M
            self.assertTrue(utils.strftime(d, "%A") in ("mercredi", "Mercredi"))

            # with text
            self.assertEqual(
                utils.strftime(d, "Écrit le %d %B %Y"), "Écrit le 29 août 2012"
            )

            # non-ascii format candidate (someone might pass it… for some reason)
            self.assertEqual(utils.strftime(d, "%écrits en %Y"), "%écrits en 2012")

    def test_maybe_pluralize(self):
        self.assertEqual(utils.maybe_pluralize(0, "Article", "Articles"), "0 Articles")
        self.assertEqual(utils.maybe_pluralize(1, "Article", "Articles"), "1 Article")
        self.assertEqual(utils.maybe_pluralize(2, "Article", "Articles"), "2 Articles")

    def test_temporary_locale(self):
        # test with default LC category
        orig_locale = locale.setlocale(locale.LC_ALL)

        with utils.temporary_locale("C"):
            self.assertEqual(locale.setlocale(locale.LC_ALL), "C")

        self.assertEqual(locale.setlocale(locale.LC_ALL), orig_locale)

        # test with custom LC category
        orig_locale = locale.setlocale(locale.LC_TIME)

        with utils.temporary_locale("C", locale.LC_TIME):
            self.assertEqual(locale.setlocale(locale.LC_TIME), "C")

        self.assertEqual(locale.setlocale(locale.LC_TIME), orig_locale)


class TestCopy(unittest.TestCase):
    """Tests the copy utility"""

    def setUp(self):
        self.root_dir = mkdtemp(prefix="pelicantests.")
        self.old_locale = locale.setlocale(locale.LC_ALL)
        locale.setlocale(locale.LC_ALL, "C")

    def tearDown(self):
        shutil.rmtree(self.root_dir)
        locale.setlocale(locale.LC_ALL, self.old_locale)

    def _create_file(self, *path):
        with open(os.path.join(self.root_dir, *path), "w") as f:
            f.write("42\n")

    def _create_dir(self, *path):
        os.makedirs(os.path.join(self.root_dir, *path))

    def _exist_file(self, *path):
        path = os.path.join(self.root_dir, *path)
        self.assertTrue(os.path.isfile(path), f"File does not exist: {path}")

    def _exist_dir(self, *path):
        path = os.path.join(self.root_dir, *path)
        self.assertTrue(os.path.exists(path), f"Directory does not exist: {path}")

    def test_copy_file_same_path(self):
        self._create_file("a.txt")
        utils.copy(
            os.path.join(self.root_dir, "a.txt"), os.path.join(self.root_dir, "b.txt")
        )
        self._exist_file("b.txt")

    def test_copy_file_different_path(self):
        self._create_dir("a")
        self._create_dir("b")
        self._create_file("a", "a.txt")
        utils.copy(
            os.path.join(self.root_dir, "a", "a.txt"),
            os.path.join(self.root_dir, "b", "b.txt"),
        )
        self._exist_dir("b")
        self._exist_file("b", "b.txt")

    def test_copy_file_create_dirs(self):
        self._create_file("a.txt")
        utils.copy(
            os.path.join(self.root_dir, "a.txt"),
            os.path.join(self.root_dir, "b0", "b1", "b2", "b3", "b.txt"),
        )
        self._exist_dir("b0")
        self._exist_dir("b0", "b1")
        self._exist_dir("b0", "b1", "b2")
        self._exist_dir("b0", "b1", "b2", "b3")
        self._exist_file("b0", "b1", "b2", "b3", "b.txt")

    def test_copy_dir_same_path(self):
        self._create_dir("a")
        self._create_file("a", "a.txt")
        utils.copy(os.path.join(self.root_dir, "a"), os.path.join(self.root_dir, "b"))
        self._exist_dir("b")
        self._exist_file("b", "a.txt")

    def test_copy_dir_different_path(self):
        self._create_dir("a0")
        self._create_dir("a0", "a1")
        self._create_file("a0", "a1", "a.txt")
        self._create_dir("b0")
        utils.copy(
            os.path.join(self.root_dir, "a0", "a1"),
            os.path.join(self.root_dir, "b0", "b1"),
        )
        self._exist_dir("b0", "b1")
        self._exist_file("b0", "b1", "a.txt")

    def test_copy_dir_create_dirs(self):
        self._create_dir("a")
        self._create_file("a", "a.txt")
        utils.copy(
            os.path.join(self.root_dir, "a"),
            os.path.join(self.root_dir, "b0", "b1", "b2", "b3", "b"),
        )
        self._exist_dir("b0")
        self._exist_dir("b0", "b1")
        self._exist_dir("b0", "b1", "b2")
        self._exist_dir("b0", "b1", "b2", "b3")
        self._exist_dir("b0", "b1", "b2", "b3", "b")
        self._exist_file("b0", "b1", "b2", "b3", "b", "a.txt")


class TestDateFormatter(unittest.TestCase):
    """Tests that the output of DateFormatter jinja filter is same as
    utils.strftime"""

    def setUp(self):
        # prepare a temp content and output folder
        self.temp_content = mkdtemp(prefix="pelicantests.")
        self.temp_output = mkdtemp(prefix="pelicantests.")

        # prepare a template file
        template_dir = os.path.join(self.temp_content, "template")
        template_path = os.path.join(template_dir, "source.html")
        os.makedirs(template_dir)
        with open(template_path, "w") as template_file:
            template_file.write('date = {{ date|strftime("%A, %d %B %Y") }}')
        self.date = utils.SafeDatetime(2012, 8, 29)

    def tearDown(self):
        shutil.rmtree(self.temp_content)
        shutil.rmtree(self.temp_output)
        # reset locale to default
        locale.setlocale(locale.LC_ALL, "")

    @unittest.skipUnless(
        locale_available("fr_FR.UTF-8") or locale_available("French"),
        "French locale needed",
    )
    def test_french_strftime(self):
        # This test tries to reproduce an issue that
        # occurred with python3.3 under macos10 only
        temp_locale = "French" if platform == "win32" else "fr_FR.UTF-8"

        with utils.temporary_locale(temp_locale):
            date = utils.SafeDatetime(2014, 8, 14)
            # we compare the lower() dates since macos10 returns
            # "Jeudi" for %A whereas linux reports "jeudi"
            self.assertEqual(
                "jeudi, 14 août 2014",
                utils.strftime(date, date_format="%A, %d %B %Y").lower(),
            )
            df = utils.DateFormatter()
            self.assertEqual(
                "jeudi, 14 août 2014", df(date, date_format="%A, %d %B %Y").lower()
            )

        # Let us now set the global locale to C:
        with utils.temporary_locale("C"):
            # DateFormatter should still work as expected
            # since it is the whole point of DateFormatter
            # (This is where pre-2014/4/15 code fails on macos10)
            df_date = df(date, date_format="%A, %d %B %Y").lower()
            self.assertEqual("jeudi, 14 août 2014", df_date)

    @unittest.skipUnless(
        locale_available("fr_FR.UTF-8") or locale_available("French"),
        "French locale needed",
    )
    def test_french_locale(self):
        if platform == "win32":
            locale_string = "French"
        else:
            locale_string = "fr_FR.UTF-8"
        settings = read_settings(
            override={
                "LOCALE": locale_string,
                "TEMPLATE_PAGES": {"template/source.html": "generated/file.html"},
            }
        )

        generator = TemplatePagesGenerator(
            {"date": self.date}, settings, self.temp_content, "", self.temp_output
        )
        generator.env.filters.update({"strftime": utils.DateFormatter()})

        writer = Writer(self.temp_output, settings=settings)
        generator.generate_output(writer)

        output_path = os.path.join(self.temp_output, "generated", "file.html")

        # output file has been generated
        self.assertTrue(os.path.exists(output_path))

        # output content is correct
        with utils.pelican_open(output_path) as output_file:
            self.assertEqual(
                output_file, utils.strftime(self.date, "date = %A, %d %B %Y")
            )

    @unittest.skipUnless(
        locale_available("tr_TR.UTF-8") or locale_available("Turkish"),
        "Turkish locale needed",
    )
    def test_turkish_locale(self):
        if platform == "win32":
            locale_string = "Turkish"
        else:
            locale_string = "tr_TR.UTF-8"
        settings = read_settings(
            override={
                "LOCALE": locale_string,
                "TEMPLATE_PAGES": {"template/source.html": "generated/file.html"},
            }
        )

        generator = TemplatePagesGenerator(
            {"date": self.date}, settings, self.temp_content, "", self.temp_output
        )
        generator.env.filters.update({"strftime": utils.DateFormatter()})

        writer = Writer(self.temp_output, settings=settings)
        generator.generate_output(writer)

        output_path = os.path.join(self.temp_output, "generated", "file.html")

        # output file has been generated
        self.assertTrue(os.path.exists(output_path))

        # output content is correct
        with utils.pelican_open(output_path) as output_file:
            self.assertEqual(
                output_file, utils.strftime(self.date, "date = %A, %d %B %Y")
            )


class TestSanitisedJoin(unittest.TestCase):
    def test_detect_parent_breakout(self):
        with self.assertRaisesRegex(
            RuntimeError,
            "Attempted to break out of output directory to (.*?:)?/foo/test",
        ):  # (.*?:)? accounts for Windows root
            utils.sanitised_join("/foo/bar", "../test")

    def test_detect_root_breakout(self):
        with self.assertRaisesRegex(
            RuntimeError,
            "Attempted to break out of output directory to (.*?:)?/test",
        ):  # (.*?:)? accounts for Windows root
            utils.sanitised_join("/foo/bar", "/test")

    def test_pass_deep_subpaths(self):
        self.assertEqual(
            utils.sanitised_join("/foo/bar", "test"),
            utils.posixize_path(os.path.abspath(os.path.join("/foo/bar", "test"))),
        )


class TestMemoized(unittest.TestCase):
    def test_memoized(self):
        class Container:
            def _get(self, key):
                pass

            @utils.memoized
            def get(self, key):
                return self._get(key)

        container = Container()

        with unittest.mock.patch.object(
            container, "_get", side_effect=lambda x: x
        ) as get_mock:
            self.assertEqual("foo", container.get("foo"))
            get_mock.assert_called_once_with("foo")

            get_mock.reset_mock()
            self.assertEqual("foo", container.get("foo"))
            get_mock.assert_not_called()

            self.assertEqual("bar", container.get("bar"))
            get_mock.assert_called_once_with("bar")

            get_mock.reset_mock()
            container.get.cache.clear()
            self.assertEqual("bar", container.get("bar"))
            get_mock.assert_called_once_with("bar")


class TestStringUtils(unittest.TestCase):
    def test_file_suffix(self):
        self.assertEqual("", utils.file_suffix(""))
        self.assertEqual("", utils.file_suffix("foo"))
        self.assertEqual("md", utils.file_suffix("foo.md"))



from blinker import Signal, signal
from ordered_set import OrderedSet

# Signals will call functions in the order of connection, i.e. plugin order
Signal.set_class = OrderedSet

# Run-level signals:

initialized = signal("pelican_initialized")
get_generators = signal("get_generators")
all_generators_finalized = signal("all_generators_finalized")
get_writer = signal("get_writer")
finalized = signal("pelican_finalized")

# Reader-level signals

readers_init = signal("readers_init")

# Generator-level signals

generator_init = signal("generator_init")

article_generator_init = signal("article_generator_init")
article_generator_pretaxonomy = signal("article_generator_pretaxonomy")
article_generator_finalized = signal("article_generator_finalized")
article_generator_write_article = signal("article_generator_write_article")
article_writer_finalized = signal("article_writer_finalized")

page_generator_init = signal("page_generator_init")
page_generator_finalized = signal("page_generator_finalized")
page_generator_write_page = signal("page_generator_write_page")
page_writer_finalized = signal("page_writer_finalized")

static_generator_init = signal("static_generator_init")
static_generator_finalized = signal("static_generator_finalized")

# Page-level signals

article_generator_preread = signal("article_generator_preread")
article_generator_context = signal("article_generator_context")

page_generator_preread = signal("page_generator_preread")
page_generator_context = signal("page_generator_context")

static_generator_preread = signal("static_generator_preread")
static_generator_context = signal("static_generator_context")

content_object_init = signal("content_object_init")

# Writers signals
content_written = signal("content_written")
feed_generated = signal("feed_generated")
feed_written = signal("feed_written")
